# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compliance automation underlying test framework utilities module."""


def parse_test_id(test_id):
    """
    Parse a test ID into useful parts.

    Takes a test ID and parses out useful parts of it::

        > parse_test_id('foo.bar.baz.test_mod.MyTestClass.test_method')
        {
            'scope': 'foo',
            'type': 'bar',
            'accreditation': 'baz',
            'file': 'test_mod',
            'class': 'MyTestClass',
            'method': 'test_method',
            'class_path': 'foo.bar.baz.test_mod.MyTestClass.test_method'
        }

    Note: scope/type/accreditation might not be returned if your
    path structure is different from the suggested one.
    """
    parts = test_id.split('.')
    full_path = len(parts) == 6
    return {
        'scope': full_path and parts[0],
        'type': full_path and parts[1],
        'accreditation': full_path and parts[2],
        'file': parts[-3],
        'class': parts[-2],
        'method': parts[-1],
        'class_path': '.'.join(parts[0:-1])
    }



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compliance automation command line interface."""

from compliance import __version__
from compliance.notify import get_notifiers
from compliance.runners import CheckMode, FetchMode

from ilcli import Command


class ComplianceCLI(Command):
    """The Compliance Framework CLI."""

    name = 'compliance'

    def _init_arguments(self):
        self.add_argument(
            '-V',
            '--version',
            help='Displays the auditree framework version.',
            action='version',
            version=f'Auditree Framework version v{__version__}'
        )
        self.add_argument(
            '-v',
            '--verbose',
            help='Displays verbose output.',
            action='store_const',
            const=2,
            default=1
        )
        self.add_argument(
            '--fetch',
            help='Enables the fetch process.',
            action='store_true',
            default=False
        )
        self.add_argument(
            '--check',
            help=(
                'Enables the check process.  '
                'Check groupings can be a comma separated list without spaces.'
            ),
            metavar='chk.grp1,chk.grp2,...',
            nargs='?',
            const=''
        )
        self.add_argument(
            '--evidence',
            help=(
                'Defines the evidence storage mode.  Defaults to %(default)s.'
            ),
            choices=['local', 'no-push', 'full-remote'],
            default='no-push'
        )
        self.add_argument(
            '--fix',
            help='Attempts to fix check failures.  Defaults to %(default)s.',
            choices=['off', 'on', 'dry-run'],
            default='off'
        )
        self.add_argument(
            '-C',
            '--compliance-config',
            help='Specifies the path/name of the compliance config JSON file.',
            metavar='auditree.json',
            default=None
        )
        self.add_argument(
            '--creds-path',
            help=(
                'Specifies the path/name of the credentials ini file.  '
                'Defaults to %(default)s.'
            ),
            metavar='/path/to/creds.ini',
            default='~/.credentials'
        )
        notify_options = [k for k in get_notifiers().keys() if k != 'stdout']
        self.add_argument(
            '--notify',
            help=(
                'Specifies a list of notifiers for sending notifications.  '
                'Valid values (can be a comma separated list - no spaces): '
                f'{", ".join(notify_options)}.  NOTE: In addition to those '
                'specified, the %(default)s notifier will always execute.'
            ),
            metavar='[slack,gh_issues,...]',
            default='stdout'
        )
        self.add_argument(
            '--force',
            help='Forces an evidence to be fetched, ignoring TTL.',
            metavar='raw/category/evidence.ext',
            action='append',
            default=[]
        )
        self.add_argument(
            '--include',
            help='Specifies the path/name of the fetcher include JSON file.',
            metavar='fetchers.json',
            default=None
        )
        self.add_argument(
            '--exclude',
            help='Specifies the path/name of the fetcher exclude JSON file.',
            metavar='fetchers.json',
            default=None
        )

    def _validate_arguments(self, args):
        if 'stdout' not in args.notify:
            args.notify += ',stdout'
        if args.check == '':
            self.parser.error(
                '--check option requires accreditation grouping(s).'
            )
        if not args.fetch and not args.check:
            self.parser.error('--fetch or --check option is expected.')
        if not args.fetch and (args.include or args.exclude):
            self.parser.error(
                '--include/--exclude options only valid with --fetch.'
            )
        if not args.check and args.fix != 'off':
            self.parser.error('--fix option only valid with --check.')

    def _validate_extra_arguments(self, extra_args):
        self.extra_args = list(set(extra_args) - {'--no-nose', '-s'})
        unrecognized = [ea for ea in self.extra_args if ea.startswith('-')]
        if unrecognized:
            self.parser.error(
                f'unrecognized arguments: {", ".join(unrecognized)}'
            )
        if '-s' in extra_args:
            self.out('WARNING: The -s option is deprecated/no longer used.')

    def _run(self, args):
        success = True
        if args.fetch:
            if args.evidence == 'full-remote':
                self.out(
                    'INFO: A remote locker sync will only occur '
                    'after checks are executed in full-remote mode.'
                )
            with FetchMode(args, self.extra_args) as fetch:
                # Handle fetcher primary run.
                self.out('\nFetcher Primary Run\n')
                success = fetch.run_fetchers()
                # Handle fetcher dependency reruns.
                previous = set()
                reruns = fetch.locker.get_dependency_reruns()
                rerun_count = 1
                while reruns and reruns != previous and rerun_count <= 100:
                    # Upper bound for reruns set to 100
                    # to guard against endless executions.
                    self.out(f'\nFetcher Dependency Re-Run #{rerun_count}\n')
                    success = fetch.run_fetchers(reruns) and success
                    rerun_count += 1
                    previous = reruns
                    reruns = fetch.locker.get_dependency_reruns()
                if reruns:
                    success = False
                    self.err(
                        '\nUnable to resolve dependency issues with %s.\n',
                        ', '.join(reruns)
                    )
                for fetch_load_error in fetch.load_errors:
                    self.err(f'\nERROR: {fetch_load_error}\n')
        if args.check:
            with CheckMode(args, self.extra_args) as check:
                accreds = ', '.join(check.accreds)
                self.out(f'\nCheck Run - Accreditations: {accreds}\n')
                success = check.run_checks() and success
                for check_load_error in check.load_errors:
                    self.err(f'\nERROR: {check_load_error}\n')
        return 0 if success else 1


def run():
    """Execute the Compliance CLI."""
    exit(ComplianceCLI().run())


if __name__ == '__main__':
    run()



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compliance report build automation module."""

import copy
import datetime
import logging
import pathlib
import traceback

from compliance.config import get_config
from compliance.evidence import get_evidence_by_path
from compliance.locker import READMES
from compliance.utils.data_parse import format_json

import jinja2

log = logging.getLogger('compliance.report')


class ReportBuilder(object):
    """This class builds all the required reports for the tests."""

    def __init__(self, locker, results, controls):
        """
        Construct and initialize the file descriptor notifier object.

        :param locker: the locker to be used.
        :param results: the dictionary of the test results.
        :param controls: the control descriptor that manages accreditations.
        """
        self.locker = locker
        self.results = results
        self.controls = controls

    def build(self):
        """Build the reports and store them in the locker."""
        test_by_class = self._get_test_by_class()
        with self.locker:
            self._generate_reports(test_by_class)
            rpt_metadata = self.locker.get_reports_metadata()
            self.generate_toc(rpt_metadata)
            self.generate_check_results(rpt_metadata)

    def render_evidence_with_template(self, evidence, test_obj):
        """
        Render content based on a template.

        If template system was selected, this method renders the content
        using the template for the given evidence.

        :param evidence: the evidence object to be rendered using its template.
        :param test_obj: test object which has all the tests for check.
        """
        if evidence.content is not None:
            return
        tmpl_path = pathlib.Path(self.get_template_for(test_obj, evidence))
        now = datetime.datetime.utcnow()
        context = {
            'test': test_obj,
            'results': self.results,
            'all_successes': test_obj.successes_for_check(self.results),
            'all_failures': test_obj.failures_for_check(self.results),
            'all_warnings': test_obj.warnings_for_check(self.results),
            'evidence': evidence,
            'builder': self,
            'now': now
        }
        loader = jinja2.FileSystemLoader(str(tmpl_path.parent))
        env = jinja2.Environment(loader=loader, autoescape=True)
        evidence.set_content(env.get_template(tmpl_path.name).render(context))

    def get_template_for(self, test_obj, evidence):
        """
        Provide the file path of the template associated to the given evidence.

        :param test_report_obj: the test object that needs to create report.
        :param evidence: the ReportEvidence object expected.
        """
        tmpl_dir = get_config().get_template_dir(test_obj)
        if tmpl_dir is None:
            raise RuntimeError(
                f'Unable to find template directory for test {test_obj.id()}'
            )
        tmpl_path = pathlib.Path(tmpl_dir, f'{evidence.path}.tmpl')
        if not tmpl_path.exists():
            tmpl_path = pathlib.Path(tmpl_dir, 'default.md.tmpl')
        return str(tmpl_path)

    def generate_toc(self, rpt_metadata):
        """
        Generate a check reports table of contents.

        This method generates a TOC based on all report evidence metadata and
        appends that TOC to the bottom of an evidence locker's README.md file.

        :param rpt_metadata: Metadata from all report evidence index.json files
        """
        path = pathlib.Path(self.locker.local_path)
        files = sorted(
            str(f) for f in path.iterdir() if f.is_file() and f.name in READMES
        )
        readme = files[0] if files else 'README.md'
        content_as_str = self.locker.get_content_from_locker(filename=readme)
        rpts = []
        for rpt, meta in rpt_metadata.items():
            if meta.get('pruned_by'):
                continue
            rpt_descr = meta['description'] or pathlib.Path(rpt).name
            rpt_url = self.locker.get_remote_location(rpt, False)
            check = meta.get('checks', ['N/A'])[0].rsplit('.', 1).pop(0)
            evidences = []
            for ev in meta.get('evidence', []):
                ev_path = pathlib.Path(ev['path'])
                ev_descr = ev['description'] or ev_path.name
                if not ev.get('partitions'):
                    ev_url = self.locker.get_remote_location(
                        ev['path'], False, ev['commit_sha']
                    )
                    evidences.append(
                        {
                            'descr': ev_descr,
                            'url': ev_url,
                            'from': ev['last_update']
                        }
                    )
                else:
                    for hash_key, part in ev['partitions'].items():
                        ev_part_descr = f'{ev_descr} - {hash_key} partition'
                        ev_url = self.locker.get_remote_location(
                            str(ev_path.parent / f'{hash_key}_{ev_path.name}'),
                            False,
                            part['commit_sha']
                        )
                        evidences.append(
                            {
                                'descr': ev_part_descr,
                                'url': ev_url,
                                'from': ev['last_update']
                            }
                        )
            accreditations = sorted(self.controls.get_accreditations(check))
            rpts.append(
                {
                    'descr': rpt_descr,
                    'url': rpt_url,
                    'check': check,
                    'accreditations': ', '.join(accreditations) or 'N/A',
                    'from': meta['last_update'],
                    'evidences': sorted(evidences, key=lambda ev: ev['descr'])
                }
            )
        context = {
            'original': content_as_str.split('\n') if content_as_str else [],
            'reports': sorted(rpts, key=lambda r: r['descr'])
        }
        loader = jinja2.FileSystemLoader(get_config().get_template_dir(self))
        env = jinja2.Environment(loader=loader, autoescape=True)
        content = env.get_template('readme_toc.md.tmpl').render(context)
        self.locker.add_content_to_locker(content, filename=readme)

    def generate_check_results(self, rpt_metadata):
        """
        Combine the check execution results with associated reports metadata.

        This method combines check results with details about associated
        reports and evidences used, found in the report metadata.  It
        returns a dictionary keyed by check class dot path.

        :param rpt_metadata: Metadata from all report evidence index.json files
        """
        chk_results = {}
        for rpt, meta in rpt_metadata.items():
            check_methods = {}
            if not meta.get('checks'):
                continue
            for check in meta['checks']:
                check_class, check_method = check.rsplit('.', 1)
                check_methods[check_method] = {}
                if self.results.get(check):
                    test = self.results[check]['test'].test
                    check_methods[check_method] = {
                        'status': self.results[check]['status'],
                        'timestamp': self.results[check]['timestamp'],
                        'warnings': test.warnings,
                        'failures': test.failures,
                        'successes': test.successes,
                        'warnings_count': test.warnings_count(),
                        'failures_count': test.failures_count(),
                        'successes_count': test.successes_count()
                    }
            if not chk_results.get(check_class):
                chk_results[check_class] = {
                    'checks': check_methods,
                    'reports': {
                        rpt: meta['description']
                    },
                    'evidence': meta['evidence'],
                    'accreditations': list(
                        self.controls.get_accreditations(check_class)
                    )
                }
            else:
                chk_results[check_class]['reports'][rpt] = meta['description']
        self.locker.add_content_to_locker(
            format_json(chk_results, skipkeys=True, default=str),
            filename='check_results.json'
        )

    def _get_test_by_class(self):
        """
        Collect one test object per ComplianceTest class.

        This is required to group tests into a single report. Note that
        we only need _one_ test object of a certain ComplianceTest and
        it its attributes get updated with each test object found for
        that class. This will be useful for templates, where a test
        object is passed and it will holds all possible attributes.
        """
        retval = {}
        for _, info in self.results.items():
            test_obj = info['test'].test
            if not hasattr(test_obj, 'get_reports'):
                continue
            test_class = test_obj.__class__
            if test_class in retval:
                retval[test_class].__dict__.update(test_obj.__dict__)
            else:
                retval[test_class] = copy.copy(test_obj)
        return retval

    def _generate_reports(self, test_by_class):
        """
        Generate the reports per-test-class basis.

        If ``test.get_reports()`` replies with a list of evidence paths,
        then the template system is used. If not, a list of evidences
        is expected.

        :param test_by_class: a dictionary of {class: test_obj}.
        """
        for test_class, test_obj in test_by_class.items():
            # get a list of all test objects related to this test class
            test_infos = [
                info for info in self.results.values()
                if info['test'].test.__class__ == test_class
            ]
            try:
                reports = test_obj.get_reports()
            except (AttributeError, ValueError) as e:
                log.warning(
                    f'\n  Failed to generate report for {test_obj.id()}'
                    f'\n  Error: {e}'
                )
                for info in test_infos:
                    info['status'] = 'error'
                continue
            for r in reports:
                try:
                    self.__render_report(r, test_obj, test_infos)
                except Exception as e:
                    log.warning(
                        f'\n  Failed to generate report for {test_obj.id()}'
                        f'\n  Error: {e.__class__.__name__} '
                        f'- {traceback.format_exc(-5)}'
                    )
                    for info in test_infos:
                        info['status'] = 'error'

    def _get_checks(self, test_obj):
        """
        Get the check paths for all checks in the test object.

        :param test_obj: test object which has all the tests for check.
        """
        checks = []
        for test_id, info in self.results.items():
            if info['test'].test == test_obj:
                checks = [
                    c for c in self.results.keys()
                    if c.startswith(test_id.rsplit('.', 1)[0])
                ]
                break
        return checks

    def __render_report(self, report, test_obj, test_infos):
        evidence = report
        if isinstance(report, str):
            path = report
            if not report.startswith('reports/'):
                path = 'reports/' + report
            evidence = get_evidence_by_path(path)
        self.render_evidence_with_template(evidence, test_obj)
        self.locker.add_evidence(
            evidence,
            self._get_checks(test_obj),
            test_obj.evidences_for_check(self.results)
        )
        for test_obj_orig in [i['test'].test for i in test_infos]:
            test_obj_orig.reports.append(evidence)



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compliance evidence management automation module."""

import inspect
import json
from collections import namedtuple
from functools import wraps
from pathlib import PurePath

from compliance.check import ComplianceCheck
from compliance.config import get_config
from compliance.utils.data_parse import format_json, parse_dot_key
from compliance.utils.exceptions import (
    DependencyFetcherNotFoundError,
    DependencyUnavailableError,
    EvidenceNotFoundError,
    StaleEvidenceError
)
from compliance.utils.path import FETCH_PREFIX, substitute_config

from deprecated import deprecated

HOUR = 60 * 60
DAY = HOUR * 24
YEAR = DAY * 365

LazyLoader = namedtuple('LazyLoader', 'path ev_class')


class _BaseEvidence(object):
    """
    The evidence base class.

    :param name: the name of the evidence (e.g. 'users.csv', or
      '{org[name]}-backups.json'). Note that this parameter can be a string
      template where config values can be used.
    :param category: the category or service where this category will be
      classified (e.g. 'dbs', 'github', etc.). This is useful for grouping
      evidence that are considered of the same type or that belong to a
      specific type of source.
    :param ttl: the time (in seconds) after which the evidence will be
      considered as stale/invalid.
    :param description: a string containing a longer description about
      the evidence.
    """

    def __init__(self, name, category, ttl=DAY, description='', **kwargs):
        self._name = name
        self.category = category
        self.ttl = ttl
        self.description = description
        self._content = None

    @classmethod
    def from_evidence(cls, evidence):
        new_evidence = cls(
            evidence.name,
            evidence.category,
            evidence.ttl,
            evidence.description
        )
        if evidence.content:
            new_evidence.set_content(evidence.content)
        return new_evidence

    @classmethod
    def lazy_load(cls, path):
        return LazyLoader(path, cls)

    @classmethod
    def from_locker(cls, path, locker):
        path = substitute_config(path)
        return cls(locker.get_evidence(path))

    @property
    def rootdir(self):
        raise NotImplementedError('The root directory is not set.')

    @property
    def dir_path(self):
        return str(PurePath(self.rootdir).joinpath(self.category))

    @property
    def name(self):
        return substitute_config(self._name)

    @property
    def path(self):
        return str(PurePath(self.dir_path).joinpath(self.name))

    @property
    def extension(self):
        return PurePath(self.name).suffix.lstrip('.')

    @property
    def content(self):
        return self._content

    @property
    def content_as_json(self):
        if self.extension != 'json':
            raise ValueError(f'{self.name} does not have JSON content.')
        if not hasattr(self, '_content_as_json'):
            self._content_as_json = json.loads(self.content)
        return self._content_as_json

    def set_content(self, str_content):
        self._content = str_content
        if self.extension == 'json':
            self._content = format_json(json.loads(str_content))


class RawEvidence(_BaseEvidence):
    """
    The raw evidence class.

    Raw evidence is partition-able if the content is JSON and by providing a
    list of partition fields along with an optional partition root.  The
    evidence partitioning can be defined in configuration JSON or directly upon
    construction of a RawEvidence object.  The constructor route overrides the
    configuration settings, if both are provided.
    """

    def __init__(self, *args, **kwargs):
        """Construct and initialize the raw evidence object."""
        super().__init__(*args, **kwargs)
        lp_config = get_config().get('locker.partitions', {})
        partition = kwargs.get(
            'partition', lp_config.get(f'{self.category}/{self.name}', {})
        )
        self.part_fields = partition.get('fields')
        self.part_root = partition.get('root')

    @property
    def rootdir(self):
        """Root directory for raw evidence."""
        return 'raw'

    @property
    def is_partitioned(self):
        """Raw evidence partitioned status."""
        return self.part_fields is not None and self.extension == 'json'

    @property
    def partition_keys(self):
        """
        All partition keys for this evidence object.

        Return all key values based on evidence content and the
        key fields provided during evidence instantiation.  Key values are
        relative to the root partition provided during evidence instantiation.

        :returns: A list of key values where the key values are lists as well.
        """
        if getattr(self, '_partition_keys', None) is None:
            keys = set()
            if self.part_root:
                root = parse_dot_key(json.loads(self._content), self.part_root)
            else:
                root = json.loads(self._content)
            for data in root:
                key_values = []
                for field in self.part_fields:
                    key_value = data
                    key_values.append(parse_dot_key(key_value, field))
                keys.add(tuple(key_values))
            self._partition_keys = [list(k) for k in keys]
        return self._partition_keys

    def get_partition(self, key):
        """
        Provide a slice of content based on the supplied partition key.

        Return a JSON document that is a slice of the original evidence
        content based on the list of key values provided.  Key values are
        expected to match the key fields provided during evidence
        instantiation and are relative to the root partition provided during
        evidence instantiation.

        :param key: A list of key values to partition by

        :returns: A JSON document filtered by the key values provided
        """
        data = json.loads(self._content)
        if not self.part_root:
            data = self._partition(data, key)
        else:
            part = data
            root = self.part_root.split('.')
            for field in root[:-1]:
                part = part[field]
            part[root[-1]] = self._partition(part[root[-1]], key)
        return format_json(data)

    def _partition(self, data, key):
        idx = 0
        for field in self.part_fields:
            data = list(
                filter(lambda e: parse_dot_key(e, field) == key[idx], data)
            )
            idx += 1
        return data


@deprecated(reason='Evidence class to be removed, use RawEvidence instead')
class DerivedEvidence(_BaseEvidence):
    """The derived evidence class."""

    @property
    def rootdir(self):
        """Root directory for derived evidence."""
        return 'derived'


class ReportEvidence(_BaseEvidence):
    """The report evidence class."""

    @property
    def rootdir(self):
        """Root directory for report evidence."""
        return 'reports'


@deprecated(reason='Evidence class to be removed')
class TmpEvidence(_BaseEvidence):
    """The temporary evidence class."""

    @property
    def rootdir(self):
        """Root directory for temporary evidence."""
        return 'tmp'


class ExternalEvidence(_BaseEvidence):
    """The external evidence class."""

    def __init__(self, name, category, ttl=YEAR, description='', **kwargs):
        """Construct and initialize the external evidence object."""
        super().__init__(name, category, ttl, description, **kwargs)

    @property
    def rootdir(self):
        """Root directory for external evidence."""
        return 'external'


class _EvidenceContextManager(object):
    """Base class for raw evidence context managers."""

    def __init__(self, locker, evidence_path, evidence_type):
        self.locker = locker
        self.evidence_path = evidence_path
        self.evidence_type = evidence_type

    def __enter__(self):
        path = self.evidence_path
        if not path.startswith(f'{self.evidence_type}/'):
            path = f'{self.evidence_type}/{path}'
        self.evidence = get_evidence_by_path(path)
        if self.locker.validate(self.evidence):
            self.evidence = None
        return self.evidence

    def __exit__(self, typ, val, traceback):
        if self.evidence and traceback is None:
            self.locker.add_evidence(self.evidence)


class raw_evidence(_EvidenceContextManager):  # noqa: N801
    """
    Helper context manager for a typical ``fetch_`` method implementation.

    Use when retrieving raw evidence by a fetcher and the name of the evidence
    is dynamic.  When the evidence name is static/known use the
    @store_raw_evidence decorator instead.

    If TTL has expired, the context manager yields the evidence object
    specified by the ``evidence_path``.  The context manager expects the
    evidence content to be retrieved and set via ``set_content`` by the calling
    ``with`` block.

    If TTL has not expired, the context manager yields ``None`` and expects the
    calling ``with`` block to exit gracefully immediately.
    """

    def __init__(self, locker, evidence_path):
        """
        Construct and initialize the raw evidence context manager.

        :param locker: The evidence locker object
        :param evidence_path: The path to the raw evidence within the evidence
          locker.  For example, ``src/my_raw_evidence.json``.
        """
        super().__init__(locker, evidence_path, 'raw')


@deprecated(reason='Context manager to be removed')
class tmp_evidence(_EvidenceContextManager):  # noqa: N801
    """
    Helper context manager for a typical ``fetch_`` method implementation.

    Use when retrieving temporary evidence by a fetcher and the name of the
    evidence is dynamic.  When the evidence name is static/known use the
    @store_tmp_evidence decorator instead.

    If the evidence is not present, the context manager yields the evidence
    object specified by the ``evidence_path``.  The context manager expects the
    evidence content to be retrieved and set via ``set_content`` by the calling
    ``with`` block.

    If the evidence is present, the context manager yields ``None`` and expects
    the calling ``with`` block to exit gracefully immediately.
    """

    def __init__(self, locker, evidence_path):
        """
        Construct and initialize the temporary evidence context manager.

        :param locker: The evidence locker object
        :param evidence_path: The path to the raw evidence within the evidence
          locker.  For example, ``src/my_tmp_evidence.json``.
        """
        super().__init__(locker, evidence_path, 'tmp')


@deprecated(reason='Context manager to be removed, use raw_evidence instead')
class derived_evidence(object):  # noqa: N801
    """
    Helper context manager for a typical ``fetch_`` method implementation.

    Use when retrieving derived evidence by a fetcher and the name of the
    evidence is dynamic.  When the evidence name is static/known use the
    @store_derived_evidence decorator instead.

    If TTL has expired on the derived evidence, the context manager yields a
    dictionary containing all of the source evidences as well as the derived
    evidence objects specified in the ``sources`` and ``target`` arguments.

    If TTL has not expired on the derived evidence, the context manager yields
    ``None`` and expects the calling ``with`` block to exit gracefully
    immediately.

    Inside the ``with`` block:

    If ``sources`` are provided as a list then the source evidence is
    accessible through the source path as the key for the yielded dictionary of
    evidence.

    If ``sources`` are provided as a dictionary where evidence short names are
    keys and evidence source paths are the values then the evidence is
    accessible through the evidence short name as the key for the yielded
    dictionary of evidence.

    If a single evidence source path is provided as a string then the source
    evidence is accessible by using 'source' as the key for the yielded
    dictionary of evidence.

    The derived evidence object is always accessible by using 'derived' as the
    key for the yielded dictionary of evidence.

    The context manager expects the derived evidence content to be generated
    and set via ``set_content`` by the calling ``with`` block.
    """

    def __init__(self, locker, sources, target):
        """
        Construct and initialize the derived evidence context manager.

        :param locker: The evidence locker object
        :param sources: The paths to the evidences within the evidence
          locker that will make up the derived evidence.  This can be a list,
          a dict or a string.
          For (list) example, ``['src/evidence.json', ...]``.
          For (dict) example, ``{'evidence1': 'src/evidence.json', ...}.
          For (str) example, ``'src/evidence.json'``.
        :param target: The path to the derived evidence within the evidence
          locker.  For example, ``src/my_derived_evidence.json``.
        """
        self.locker = locker
        self.sources = sources
        self.target = target
        self.evidences = None

    def __enter__(self):
        """Perform fetcher derived evidence pre-processing."""
        target_path = self.target
        if not target_path.startswith('derived/'):
            target_path = f'derived/{target_path}'
        target_evidence = get_evidence_by_path(target_path)
        if self.locker.validate(target_evidence):
            return
        self.evidences = {'derived': target_evidence}
        if isinstance(self.sources, list):
            for src in self.sources:
                self.evidences[src] = get_evidence_by_path(src, self.locker)
        elif isinstance(self.sources, dict):
            for src_name, src_path in self.sources.items():
                self.evidences[src_name] = get_evidence_by_path(
                    src_path, self.locker
                )
        elif isinstance(self.sources, str):
            self.evidences['source'] = get_evidence_by_path(
                self.sources, self.locker
            )
        else:
            self.evidences = None
        return self.evidences

    def __exit__(self, typ, val, traceback):
        """Perform fetcher derived evidence post-processing."""
        if self.evidences and traceback is None:
            self.locker.add_evidence(self.evidences['derived'])


class evidences(object):  # noqa: N801
    """
    Helper context manager for a typical ``test_`` method implementation.

    Use when processing evidence by a check and the name(s) of the
    evidence is/are dynamic.  When the evidence names are static/known use the
    @with_*_evidences decorators instead.

    Inside the ``with`` block:

    If ``evidence_paths`` are provided as a list then the evidence is
    accessible through the evidence path as the key for the yielded dictionary
    of evidence.

    If ``evidence_paths`` are provided as a dictionary where evidence short
    names are keys and evidence paths are the values then the evidence is
    accessible through the evidence short name as the key for the yielded
    dictionary of evidence.

    If a single evidence path is provided as a string then the evidence object
    is yielded by this context manager.
    """

    def __init__(self, obj, from_evidences):
        """
        Construct and initialize the evidences context manager.

        :param obj: Either a check or locker object.  Needed for backward
          compatibility of evidences context manager.
        :param from_evidences: The paths to evidences within the evidence
          locker.  It can be any of the following:
          - A string path
          - A LazyLoader namedtuple
          - A list of string paths or LazyLoader namedtuples or a
          combination of string paths and LazyLoader namedtuples
          - A dictionary where a key is an evidence short name and a value
          is either string path or a LazyLoader namedtuple

          Using the LazyLoader namedtuple tells this context manager to cast
          your evidence as a subclass of one of the framework's base evidence
          classes.
        """
        if isinstance(obj, ComplianceCheck):
            self.check = obj
            self.locker = self.check.locker
        else:
            self.locker = obj
        self.from_evidences = from_evidences

    def __enter__(self):
        """Perform check evidences pre-processing."""
        evidence = {}
        evidence_list = []
        if isinstance(self.from_evidences, list):
            for from_evidence in self.from_evidences:
                path = from_evidence
                if isinstance(from_evidence, LazyLoader):
                    # preserve original path to be used as key of evidence dict
                    path = from_evidence.path
                ev = self._get_evidence(from_evidence)
                evidence[path] = ev
                evidence_list.append(ev.path)
        elif isinstance(self.from_evidences, dict):
            for evidence_name, from_evidence in self.from_evidences.items():
                evidence[evidence_name] = self._get_evidence(from_evidence)
                evidence_list.append(evidence[evidence_name].path)
        else:
            evidence = self._get_evidence(self.from_evidences)
            evidence_list.append(evidence.path)
        if hasattr(self, 'check'):
            for ev_path in evidence_list:
                self.check.add_evidence_metadata(ev_path)
        return evidence

    def __exit__(self, typ, val, traceback):
        """Perform check evidences post-processing."""
        pass

    def _get_evidence(self, from_evidence):
        """
        Retrieve an evidence instance based on the from_evidence provided.

        from_evidence can be either a path to the evidence in the locker as a
        string or a LazyLoader namedtuple object that contains the path as a
        string and an evidence class to cast the evidence as.
        """
        path = from_evidence
        ev_class = None
        if isinstance(from_evidence, LazyLoader):
            path = from_evidence.path
            ev_class = from_evidence.ev_class
            ev_types = get_evidence_types()
            if not any(PurePath(path).parts[0] == et for et in ev_types):
                for ev_type in ev_types:
                    base_evidence_class = get_evidence_class(ev_type)
                    if ev_class and issubclass(ev_class, base_evidence_class):
                        path = str(PurePath(ev_type, path))
                        break
        evidence = get_evidence_by_path(path, self.locker)
        if evidence.__class__ not in get_evidence_base_classes():
            # Evidence returned from cache already cast
            ev_class = evidence.__class__
        base_evidence_class = get_evidence_class(evidence.rootdir)
        if not ev_class or not issubclass(ev_class, base_evidence_class):
            ev_class = base_evidence_class
        return ev_class.from_evidence(evidence)


__init_map = {
    'tmp': TmpEvidence,
    'reports': ReportEvidence,
    'derived': DerivedEvidence,
    'raw': RawEvidence,
    'external': ExternalEvidence
}


def get_evidence_class(evidence_type):
    """
    Provide the appropriate evidence class based on the supplied type.

    Returns the corresponding evidence class object based on the category
    provided.  If no match ``None`` is returned.

    :param evidence_type: the type of evidence class desired as a string. Valid
      values are ``reports``, ``raw``.

    :returns: the appropriate evidence class.
    """
    return __init_map.get(evidence_type)


def get_evidence_types():
    """Provide a list of all valid evidence types."""
    return list(__init_map.keys())


def get_evidence_base_classes():
    """Provide a list of all valid base evidence class objects."""
    return list(__init_map.values())


def get_evidence_by_path(path, locker=None):
    """
    Provide an evidence object specified by the given path.

    The following strategy is used:

      * Return evidence if it is present in the evidence cache populating
        content if necessary and possible.
      * If evidence is not in the evidence cache but a locker is provided, then
        the evidence will be retrieved from the locker.
      * Otherwise, an evidence object is built with the default parameters.

    :param path: relative path to the evidence within the Locker. For example,
      ``raw/source1/evidence.json``.

    :returns: the evidence object.
    """
    path = substitute_config(path)
    evidence = get_config().get_evidence(path)
    if evidence:
        if locker and evidence.content is None:
            evidence = locker.load_content(evidence)
        return evidence

    if locker:
        try:
            evidence = locker.get_evidence(path)
            get_config().add_evidences([evidence])
            return evidence
        except ValueError:
            pass

    try:
        rootdir, category, name = path.strip('/').split('/')
    except ValueError:
        raise ValueError(f'Invalid evidence path format "{path}"')

    if rootdir not in __init_map:
        raise ValueError(f'Unable to create evidence from root dir {rootdir}')
    return __init_map[rootdir](name, category)


def get_evidence_dependency(path, locker, fetcher=None):
    """
    Provide evidence to fetchers that depend on that other evidence to fetch.

    Use when a fetcher needs evidence fetched by another fetcher in order to
    complete its fetch process.  The premise is that if the evidence is in the
    evidence cache then return that because it was placed there by another
    fetcher.  If not then get the evidence directly from the locker without
    putting that evidence into the evidence cache.  When an evidence dependency
    is not found the fetcher is queued up as a dependency rerun for subsequent
    processing.

    :param path: relative path to the evidence within the Locker. For example,
      ``raw/source1/evidence.json``.
    :param locker: evidence Locker object.
    :param fetcher: optional Python notation path to fetcher method.  If
      provided, this defines the fetcher that is added to the re-run list.
      Otherwise the execution stack is traversed to find the fetcher caller.

    :returns: the evidence object.
    """
    path = substitute_config(path)
    evidence = get_config().get_evidence(path)
    if evidence is None or evidence.content is None:
        try:
            evidence = locker.get_evidence(path)
        except (StaleEvidenceError, EvidenceNotFoundError):
            rerun = None
            if fetcher:
                module, clss, method = fetcher.rsplit('.', 2)
                rerun = {'module': module, 'class': clss, 'method': method}
            else:
                for frame_info in inspect.stack()[1:]:
                    if frame_info.function.startswith(FETCH_PREFIX):
                        frame = frame_info.frame
                        rerun = {
                            'module': inspect.getmodule(frame).__name__,
                            'class': frame.f_locals['self'].__class__.__name__,
                            'method': frame.f_code.co_name
                        }
                        break
            if not rerun:
                raise DependencyFetcherNotFoundError(
                    f'Cannot re-run, no fetcher found for evidence {path}.'
                )
            locker.dependency_rerun.append(rerun)
            raise DependencyUnavailableError(
                f'evidence dependency {path} is currently unavailable.'
            )
    return evidence


def store_raw_evidence(evidence_path):
    """
    Decorate a typical ``fetcher_`` method fetching raw evidence.

    Use when retrieving raw evidence by a fetcher and the name of the evidence
    is static/known.  When the evidence name is dynamic use the
    raw_evidence context manager instead.

    The decorator expects that the decorated method returns the content to be
    stored in the locker.  The storing of the evidence is also handled by this
    decorator.

    :param path: relative path to the evidence within the Locker. For example,
      ``source1/evidence.json``.
    """

    def decorator(f):

        @wraps(f)  # required for preserving the function context
        def wrapper(self, *args, **kwargs):
            return _store_wrapper(self, evidence_path, f, 'raw')

        return wrapper

    return decorator


@deprecated(reason='Decorator to be removed')
def store_tmp_evidence(evidence_path):
    """
    Decorate a typical ``fetcher_`` method fetching temporary evidence.

    Use when retrieving temporary evidence by a fetcher and the name of the
    evidence is static/known.  When the evidence name is dynamic use the
    tmp_evidence context manager instead.

    :param path: relative path to the evidence within the Locker. For example,
      ``source1/evidence.json``.
    """

    def decorator(f):

        @wraps(f)  # required for preserving the function context
        def wrapper(self, *args, **kwargs):
            return _store_wrapper(self, evidence_path, f, 'tmp')

        return wrapper

    return decorator


@deprecated(reason='Decorator to be removed, use store_raw_evidence instead')
def store_derived_evidence(evidences, target):
    """
    Decorate a typical ``fetcher_`` method fetching derived evidence.

    Use when retrieving derived evidence by a fetcher and the name of the
    evidence is static/known.  When the evidence name is dynamic use the
    tmp_evidence context manager instead.

    The decorator expects that the decorated method returns the content to be
    stored in the locker.  The storing of the evidence is also handled by this
    decorator.

    :param evidences: a list of relative paths to the evidences needed by
      the fetcher. They will be passed to the method call.
      For example, ``[raw/src/foo.json, derived/src/bar.json]``.
    :param target: relative path to the evidence to be stored in the Locker.
      For example, ``src/evidence.json``.
    """

    def decorator(f):

        @wraps(f)  # required for preserving the function context
        def wrapper(self, *args, **kwargs):
            target_path = target
            if not target_path.startswith('derived/'):
                target_path = f'derived/{target_path}'
            target_evidence = get_evidence_by_path(target_path)
            if self.locker.validate(target_evidence):
                return
            depends = [get_evidence_by_path(e, self.locker) for e in evidences]
            content = f(self, *depends)
            target_evidence.set_content(content)
            self.locker.add_evidence(target_evidence)

        return wrapper

    return decorator


def with_raw_evidences(*from_evidences):
    """
    Decorate a typical ``test_`` check method processing raw evidences.

    Use when processing raw evidence by a check and the name(s) of the
    evidence is/are static/known.  When the evidence names are dynamic use the
    ``evidences`` context manager instead.

    :param from_evidences: relative paths to evidences as strings or
      LazyLoader namedtuples that contain relative paths to evidences and
      evidence classes required by the check.
    """

    def decorator(f):
        return _with_evidence_decorator(from_evidences, f, 'raw')

    return decorator


def with_external_evidences(*from_evidences):
    """
    Decorate a typical ``test_`` check method processing external evidences.

    Use when processing external evidence by a check and the name(s) of the
    evidence is/are static/known.  When the evidence names are dynamic use the
    ``evidences`` context manager instead.

    :param from_evidences: relative paths to evidences as strings or
      LazyLoader namedtuples that contain relative paths to evidences and
      evidence classes required by the check.
    """

    def decorator(f):
        return _with_evidence_decorator(from_evidences, f, 'external')

    return decorator


@deprecated(reason='Decorator to be removed, use with_raw_evidences instead')
def with_derived_evidences(*from_evidences):
    """
    Decorate a typical ``test_`` check method processing temporary evidences.

    Use when processing derived evidence by a check and the name(s) of the
    evidence is/are static/known.  When the evidence names are dynamic use the
    ``evidences`` context manager instead.

    :param from_evidences: relative paths to evidences as strings or
      LazyLoader namedtuples that contain relative paths to evidences and
      evidence classes required by the check.
    """

    def decorator(f):
        return _with_evidence_decorator(from_evidences, f, 'derived')

    return decorator


@deprecated(reason='Decorator to be removed')
def with_tmp_evidences(*from_evidences):
    """
    Decorate a typical ``test_`` check method processing derived evidences.

    Use when processing temporary evidence by a check and the name(s) of the
    evidence is/are static/known.  When the evidence names are dynamic use the
    ``evidences`` context manager instead.

    :param from_evidences: relative paths to evidences as strings or
      LazyLoader namedtuples that contain relative paths to evidences and
      evidence classes required by the check.
    """

    def decorator(f):
        return _with_evidence_decorator(from_evidences, f, 'tmp')

    return decorator


def _store_wrapper(self, evidence_path, func, type_name):
    path = evidence_path
    if not path.startswith(f'{type_name}/'):
        path = f'{type_name}/{path}'
    evidence = get_evidence_by_path(path)
    if self.locker.validate(evidence):
        return
    content = func(self)
    evidence.set_content(content)
    self.locker.add_evidence(evidence)


def _with_evidence_decorator(from_evidences, f, type_str):
    from_evs = []
    for from_ev in from_evidences:
        path = from_ev
        # Default the evidence class to evidence of type type_str.
        # If from_ev is a LazyLoader and its evidence class is of the type
        # type_str then set the evidence class to the from_ev evidence class.
        ev_class = get_evidence_class(type_str)
        if isinstance(from_ev, LazyLoader):
            path = from_ev.path
            if issubclass(from_ev.ev_class, ev_class):
                ev_class = from_ev.ev_class
        # Ensure path is the full relative path including the type root.
        if PurePath(path).parts[0] != type_str:
            path = str(PurePath(type_str, path))
        from_evs.append(LazyLoader(path, ev_class))

    if hasattr(f, 'args'):
        f.args.extend(from_evs)
        return f

    @wraps(f)  # required for preserving the function context
    def wrapper(self, *args, **kwargs):
        return _evidence_wrapper(self, wrapper.args, f)

    wrapper.args = from_evs
    return wrapper


def _evidence_wrapper(self, from_evidences, func):
    evidences = []
    for fe in from_evidences:
        ev = get_evidence_by_path(fe.path, self.locker)
        if ev.__class__ in get_evidence_base_classes():
            # Only cast if evidence retrieved is a base evidence class object
            ev = fe.ev_class.from_evidence(ev)
        evidences.append(ev)
    for evidence in evidences:
        self.add_evidence_metadata(evidence.path)
    return func(self, *evidences)



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compliance automation path formatting utilities module."""

import imp
import os
import sys

from compliance.config import get_config

FETCH_PREFIX = 'fetch_'
CHECK_PREFIX = 'test_'


def get_toplevel_dirpath(path):
    """
    Provide the toplevel directory for the given path.

    The toplevel directory will be the one containing ``controls.json`` file.
    This function returns ``None`` if toplevel path could not be calculated.

    :param path: absolute or relative path to file or dir.
    """
    if path == '/' or path is None:
        return None
    if os.path.exists(os.path.join(path, 'controls.json')):
        return path
    return get_toplevel_dirpath(os.path.dirname(path))


def load_evidences_modules(toplevel):
    """
    Load the all evidences modules found at toplevel directory.

    This function prevents double loading.

    :param toplevel: path to the toplevel directory.
    """
    for root, dirs, _ in os.walk(toplevel):
        for d in dirs:
            try:
                mod_data = imp.find_module(
                    'evidences', [os.path.join(root, d)]
                )
            except ImportError:
                continue
            module_name = f'{d}.evidences'
            if module_name in sys.modules:
                continue
            imp.load_module(module_name, *mod_data)


def substitute_config(path_tmpl):
    """
    Substitue the config values on the given path template.

    :param path_tmpl: a string template of a path
    """
    return path_tmpl.format(**(get_config().raw_config))



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compliance locker management automation module."""

import json
import logging
import os
import re
import shutil
import tempfile
import time
from datetime import datetime as dt
from threading import Lock
from urllib.parse import urlparse

from compliance.evidence import TmpEvidence, get_evidence_class
from compliance.utils.data_parse import (
    format_json, get_sha256_hash, parse_dot_key
)
from compliance.utils.exceptions import (
    EvidenceNotFoundError,
    HistoricalEvidenceNotFoundError,
    LockerPushError,
    StaleEvidenceError
)

import git

INDEX_FILE = 'index.json'
DAY = 60 * 60 * 24
AE_DEFAULT = 30 * DAY
READMES = ['README.md', 'readme.md', 'Readme.md']
AE_EXEMPT = [INDEX_FILE] + READMES


class Locker(object):
    """
    The locker (context manager) class.

    The evidence locker maintains a git repository of evidence and places new
    evidence into it. It can validate a piece of evidence against a time to
    live.  The Locker is a context manager. On instantiation it will retrieve
    the configured git repository.
    """

    lock = Lock()

    def __init__(
        self,
        name=None,
        repo_url=None,
        branch='master',
        creds=None,
        do_push=False,
        ttl_tolerance=0,
        gitconfig=None
    ):
        """
        Construct and initialize the locker object.

        :param name: name (not path) to create a git repo in, if
          a url is not specified.
        :param repo_url: the URL of a git repository, credentials are supplied
          via envvars. If not set, create a local repo and commit without
          pushing to a remote.
        :param branch: branch of the repo, defaults to master, only used if
          repo is cloned from repo_url.
        :param creds: a compliance.utils.credentials.Config object containing
          remote locker credentials.
        :param do_push: if True, perform push to remove git repo if needed.
        :param ttl_tolerance: the applied to evidence time to live.  If within
          that tolerance the evidence will be treated as stale.
        :param gitconfig: the git configuration to be applied to the locker.
        """
        self.repo_url = repo_url
        self.repo_url_with_creds = self.repo_url
        if creds is not None:
            service_hostname = urlparse(self.repo_url).hostname
            token = None
            if 'github.com' in service_hostname:
                token = creds['github'].token
            elif 'github' in service_hostname:
                token = creds['github_enterprise'].token
            elif 'bitbucket' in service_hostname:
                token = creds['bitbucket'].token
            elif 'gitlab' in service_hostname:
                token = creds['gitlab'].token
            if token:
                self.repo_url_with_creds = re.sub(
                    '://', f'://{token}@', self.repo_url_with_creds
                )
        self.branch = branch
        if name is not None:
            self.name = name
        elif repo_url is not None and name is None:
            self.name = os.path.split(self.repo_url)[1]
        elif repo_url is None and name is None:
            self.name = 'example'
        self.local_path = os.path.join(tempfile.gettempdir(), self.name)
        self._do_push = do_push
        self.ttl_tolerance = ttl_tolerance
        self.gitconfig = gitconfig or {}
        self.logger = logging.getLogger(name='compliance.locker')
        self._handler = logging.StreamHandler()
        self._handler.setFormatter(
            logging.Formatter('%(levelname)s: %(message)s')
        )
        self.logger.handlers.clear()
        self.logger.addHandler(self._handler)
        self.logger.setLevel(logging.INFO)
        self.commit_date = dt.utcnow().isoformat()
        self.forced_evidence = []
        self.dependency_rerun = []

    @property
    def touched_files(self):
        """
        Provide paths to files that have been touched in the local locker.

        Files are considered "touched" if they have been added/updated in
        the locker.

        :returns: a list of touched files.
        """
        index_files = [
            f[0]
            for f in self.repo.index.entries.keys()
            if is_index_file(f[0])
        ]
        touched_files = []
        for f in index_files:
            index_dir = os.path.dirname(f)
            content = json.loads(open(os.path.join(self.local_path, f)).read())
            for t, desc in content.items():
                if type(desc) is dict:
                    ts = desc.get('last_update')
                else:
                    ts = desc
                if ts == self.commit_date:
                    touched_files.append(os.path.join(index_dir, t))
        return touched_files

    def reset_depenency_rerun(self):
        """Reset the dependency_rerun attribute to an empty list."""
        self.dependency_rerun = []

    def get_dependency_reruns(self):
        """
        Provide dot notation path to fetcher methods to be re-executed.

        Fetchers will only be considered for re-execution if they initially
        failed to execute due to a dependency error.

        :returns: a list of fetcher methods to re-execute.
        """
        return {
            f'{r["module"]}.{r["class"]}.{r["method"]}'
            for r in self.dependency_rerun
        }

    def init(self):
        """
        Initialize the local git repository.

        If repo_url was provided to the locker object, the repository will
        be cloned. If not, a fresh new git repository will be created locally.
        """
        if self.repo_url_with_creds:
            self.checkout()
        else:
            self._repo_init()

    def logger_init_msgs(self):
        """Log locker initialization information."""
        gpgsign = self.gitconfig.get('commit', {}).get('gpgsign')
        if self._do_push and not gpgsign:
            self.logger.warning(
                'Commits may not be cryptographically signed, best '
                'practice is to set gitconfig.commit.gpgsign to true '
                'in your locker configuration.'
            )

    def add_evidence(self, evidence, checks=None, evidence_used=None):
        """
        Add the evidence to the locker.

        The locker will generate the file internally and add it to the
        local git repository.

        :param evidence: the evidence to add. Note that evidence content
          attribute must be populated.
        :param checks: A list of checks used to generate report content.  Only
          applicable for check generated ReportEvidence.
        :param evidence_used: metadata for evidence used by a check.  Only
          applicable for check generated ReportEvidence.
        """
        if evidence.content is None:
            raise ValueError(f'Evidence {evidence.name} has no content')

        if not os.path.isdir(os.path.join(self.local_path, evidence.dir_path)):
            os.makedirs(os.path.join(self.local_path, evidence.dir_path))
        if getattr(evidence, 'is_partitioned', False):
            for key in evidence.partition_keys:
                full_path = os.path.join(
                    self.local_path,
                    evidence.dir_path,
                    f'{get_sha256_hash(key, 10)}_{evidence.name}'
                )
                with open(full_path, 'w') as f:
                    f.write(evidence.get_partition(key))
        else:
            full_path = os.path.join(self.local_path, evidence.path)
            with open(full_path, 'w') as f:
                f.write(evidence.content)
        if not evidence.path.startswith('tmp/'):
            self.index(evidence, checks, evidence_used)

    def index(self, evidence, checks=None, evidence_used=None):
        """
        Add an evidence to the git index.

        :param evidence: the evidence object.
        :param checks: A list of checks used to generate report content.  Only
          applicable for check generated ReportEvidence.
        :param evidence_used: metadata for evidence used by a check.  Only
          applicable for check generated ReportEvidence.
        """
        with self.lock:
            index_file = self.get_index_file(evidence)
            repo_files = [index_file]
            if not os.path.exists(index_file):
                metadata = {}
            else:
                metadata = json.loads(open(index_file).read())
            ev_meta = metadata.get(evidence.name, {})
            old_parts = ev_meta.get('partitions', {}).keys()
            metadata[evidence.name] = {
                'last_update': self.commit_date,
                'ttl': evidence.ttl,
                'description': evidence.description
            }
            tombstones = None
            if getattr(evidence, 'is_partitioned', False):
                unpartitioned = self.get_file(evidence.path)
                if os.path.isfile(unpartitioned):
                    # Remove/tombstone unpartitioned evidence file
                    # replaced by partitioned evidence files
                    self.repo.index.remove([unpartitioned], working_tree=True)
                    tombstones = self.create_tombstone_metadata(
                        evidence.name, ev_meta, 'Evidence is partitioned'
                    )
                parts = {}
                for key in evidence.partition_keys:
                    sha256_hash = get_sha256_hash(key, 10)
                    parts[sha256_hash] = key
                    repo_file = self.get_file(
                        f'{evidence.dir_path}/{sha256_hash}_{evidence.name}'
                    )
                    repo_files.append(repo_file)
                dead_parts = set(old_parts) - set(parts.keys())
                if dead_parts:
                    # Remove/tombstone partitioned evidence files
                    # no longer part of the evidence content
                    self.remove_partitions(evidence, dead_parts)
                    tombstones = self.create_tombstone_metadata(
                        dead_parts,
                        ev_meta,
                        'Partition no longer part of evidence'
                    )
                metadata[evidence.name].update(
                    {
                        'partition_fields': evidence.part_fields,
                        'partition_root': evidence.part_root,
                        'partitions': parts
                    }
                )
                if tombstones is None:
                    # Preserve prior tombstones
                    tombstones = ev_meta.get('tombstones')
            else:
                # Remove/tombstone partitioned evidence files
                # replaced by unpartitioned evidence file
                self.remove_partitions(evidence, old_parts)
                tombstones = self.create_tombstone_metadata(
                    old_parts, ev_meta, 'Evidence no longer partitioned'
                )
                repo_files.append(self.get_file(evidence.path))
            if tombstones:
                metadata[evidence.name]['tombstones'] = tombstones
            if checks is not None:
                metadata[evidence.name]['checks'] = checks
            if evidence_used is not None:
                metadata[evidence.name]['evidence'] = evidence_used
            with open(index_file, 'w') as f:
                f.write(format_json(metadata))
            self.repo.index.add(repo_files)

    def remove_partitions(self, evidence, hashes):
        """
        Remove partition files for the specified evidence hash keys.

        Used when switching from partitioned evidence to unpartitioned evidence
        and when partitioned file content is no longer part of the overall
        evidence content.

        :param evidence: the evidence object.
        :param hashes: an iterable with evidence partition hashes.
        """
        if not hashes:
            return
        self.repo.index.remove(
            [
                os.path.join(
                    self.local_path,
                    evidence.dir_path,
                    f'{part_hash}_{evidence.name}'
                ) for part_hash in hashes
            ],
            working_tree=True
        )

    def create_tombstone_metadata(self, candidate, metadata, reason):
        """
        Create tombstones for evidence being removed from the locker.

        :param candidate: either the name of the evidence file or an iterable
          of partition hashes.
        :param metadata: the evidence metadata dictionary.
        :param reason: the reason evidence is being removed.

        :returns: a list of tombstones.
        """
        tombstones = metadata.get('tombstones', {})
        if isinstance(candidate, str):
            tombstones.setdefault(candidate, []).append(
                {
                    'eol': self.commit_date,
                    'last_update': metadata['last_update'],
                    'reason': reason
                }
            )
        else:
            for part in candidate:
                tombstones.setdefault(part, []).append(
                    {
                        'eol': self.commit_date,
                        'last_update': metadata['last_update'],
                        'partition_fields': metadata['partition_fields'],
                        'partition_root': metadata['partition_root'],
                        'partition_key': metadata['partitions'][part],
                        'reason': reason
                    }
                )
        return tombstones

    def write_pkg_indexes(self):
        """
        Add package index files to the local git repository index.

        Update the package index files with the new timestamp for the
        updated evidence.
        """
        self.repo.index.add(
            [
                f[0]
                for f in self.repo.index.entries.keys()
                if is_index_file(f[0])
            ]
        )

    def get_index_file(self, evidence):
        """
        Provide the full path to the index file of the given evidence.

        :param evidence: the evidence object to be used.
        """
        return self.get_index_file_by_path(evidence.path)

    def get_index_file_by_path(self, evidence_path):
        """
        Provide the full path to the supplied evidence's index file.

        :param evidence_path: the path to the evidence within the evidence
          locker.

        :returns: the full path to the evidence's index file.
        """
        path = evidence_path.rsplit('/', 1).pop(0)
        return os.path.join(self.local_path, path, INDEX_FILE)

    def get_file(self, filename):
        """
        Provide the path for a file in the locker.

        The file may or may not exist in the locker.

        :param filename: the name of a file in the locker.

        :returns: the path to the filename provided.
        """
        return os.path.join(self.local_path, filename)

    def get_remote_location(self, filename, include_commit=True, sha=None):
        """
        Provide the path for a file/commit SHA in the locker.

        The file may or may not exist in the locker.

        :param filename: the name of a file in the locker.
        :param include_commit: if the latest commit SHA should be included.
        :param sha: use this commit SHA; requires include_commit to be False.

        :returns: the remote repository path to the filename provided.
        """
        if not self.repo_url_with_creds:
            return os.path.join(self.local_path, filename)

        ref = self.branch
        if include_commit:
            ref = self.repo.head.commit.hexsha
        elif not include_commit and sha:
            ref = sha
        return f'{self.repo_url}/blob/{ref}/{filename.strip("/")}'

    def get_evidence(self, evidence_path, ignore_ttl=False, evidence_dt=None):
        """
        Provide the evidence from the locker.

        The evidence may or may not exist in the locker.

        :param evidence_path: string path of the evidence within the locker.
          For example, `raw/service1/output.json`

        :returns: the populated evidence object.
        """
        evidence = None
        try:
            metadata = self.get_evidence_metadata(evidence_path, evidence_dt)
            class_type, category, evidence_name = evidence_path.split('/')
            evidence_class_obj = get_evidence_class(class_type)
            evidence = evidence_class_obj(
                evidence_name,
                category,
                metadata['ttl'],
                metadata['description'],
                partition={
                    'fields': metadata.get('partition_fields'),
                    'root': metadata.get('partition_root')
                }
            )
        except TypeError:
            raise EvidenceNotFoundError(
                f'Evidence {evidence_path} is not valid '
                'or not present within the locker'
            )
        return self.load_content(evidence, ignore_ttl, evidence_dt)

    def load_content(self, evidence, ignore_ttl=False, evidence_dt=None):
        """
        Populate the content of the evidence from the locker.

        :param evidence: an evidence object.
        :param ignore_ttl: Boolean for TTL validation.  Defaults to False.
        :param evidence_dt: The date of the evidence content to load.

        :returns: The evidence object with content.
        """
        self._validate_evidence(evidence, ignore_ttl)
        if getattr(evidence, 'is_partitioned', False):
            metadata = self.get_evidence_metadata(evidence.path, evidence_dt)
            content = None
            for part_hash in metadata['partitions'].keys():
                data = json.loads(
                    self._get_file_content(
                        f'{evidence.dir_path}/{part_hash}_{evidence.name}',
                        evidence_dt
                    )
                )
                if content is None:
                    content = data
                    root = content
                    if evidence.part_root:
                        root = parse_dot_key(root, evidence.part_root)
                    continue
                if evidence.part_root:
                    root.extend(parse_dot_key(data, evidence.part_root))
                else:
                    root.extend(data)
            evidence.set_content(format_json(content))
        else:
            evidence.set_content(
                self._get_file_content(evidence.path, evidence_dt)
            )
        return evidence

    def validate(self, evidence, ignore_ttl=False):
        """
        Validate the evidence against the time to live.

        :param evidence: an Evidence object.
        :param ignore_ttl: ignore time to live if set to True.

        :returns: True/False time to live validation.
        """
        if isinstance(evidence, TmpEvidence):
            full_path = os.path.join(self.local_path, evidence.path)
            return os.path.exists(full_path)
        try:
            self._validate_evidence(evidence, ignore_ttl)
            return True
        except (ValueError, KeyError):
            pass
        except StaleEvidenceError as stale_evidence:
            self._stale_message(stale_evidence.args[0])
        return False

    def checkout(self):
        """Pull (clone) the remote repository to the local git repository."""
        if os.path.isdir(os.path.join(self.local_path, '.git')):
            self.logger.info(f'Using locker found in {self.local_path}...')
            self.repo = git.Repo(self.local_path)
        else:
            self.logger.info(
                f'Cloning locker {self.repo_url} to {self.local_path}...'
            )
            self.repo = git.Repo.clone_from(
                self.repo_url_with_creds, self.local_path, branch=self.branch
            )
        self.init_config()

    def init_config(self):
        """Apply the git configuration."""
        with self.repo.config_writer() as cw:
            for (section, cfg) in self.gitconfig.items():
                for key, value in cfg.items():
                    cw.set_value(section, key, value)

    def checkin(self, message=None):
        """
        Commit changed files to the local git repository.

        :param message: the message to apply on the commit.
        """
        self.write_pkg_indexes()
        if not message:
            updated_files = '\n'.join(self.touched_files)
            message = (
                f'Files updated at local time {time.ctime(time.time())}'
                f'\n\n{updated_files}'
            )
        self.logger.info(
            f'Committing changes to local locker in {self.local_path}...'
        )
        try:
            diff = self.repo.index.diff('HEAD')
            if len(diff) > 0:
                self.repo.git.commit(message=message)
        except git.BadName:
            self.repo.git.commit(message=message)

    def push(self):
        """Push the local git repository to the remote repository."""
        if self._do_push:
            self.logger.info(
                f'Pushing local locker to remote repo {self.repo_url}...'
            )
            push_info = self.repo.remotes[0].push()[0]
            if push_info.flags >= git.remote.PushInfo.ERROR:
                raise LockerPushError(push_info)

    def add_content_to_locker(self, content, folder='', filename=None):
        """
        Add non-evidence related content to the local locker.

        :param content: the content to add to the local locker.
        :param folder: the folder in the local locker to add the content to.
        :param filename: the name of the file in the local locker.
        """
        path = os.path.join(self.local_path, folder)
        if not os.path.exists(path):
            os.mkdir(path)
        if not filename:
            raise ValueError('Filename cannot be empty.')
        results_file = os.path.join(path, filename)
        with open(results_file, 'w+') as f:
            f.write(content)
        self.repo.index.add([results_file])

    def get_content_from_locker(self, folder='', filename=None):
        """
        Read non-evidence related content from the local locker.

        :param folder: the folder in the local locker to get the content from.
        :param filename: the name of the file in the local locker.
        """
        if not filename:
            raise ValueError('Filename cannot be empty.')
        file_path = os.path.join(self.local_path, folder, filename)
        content = None
        if os.path.exists(file_path):
            with open(file_path) as f:
                content = f.read()
        return content

    def get_reports_metadata(self):
        """
        Provide all metadata related to reports as a dictionary.

        :returns: reports metadata.
        """
        metadata = {}
        rpts = os.walk(os.path.join(self.local_path, 'reports'))
        for path in [rpath for rpath, _, files in rpts if INDEX_FILE in files]:
            with open(os.path.join(path, INDEX_FILE), 'r') as f:
                for rpt_name, rpt_metadata in json.loads(f.read()).items():
                    rpt_path = os.path.join(
                        path.split(self.local_path).pop(), rpt_name
                    )
                    metadata[rpt_path] = rpt_metadata
        return metadata

    def get_abandoned_evidences(self, threshold=None):
        """
        Provide a list of evidence where the update ``threshold`` has passed.

        :param int threshold: the time in seconds after TTL expires that
          evidence can remain un-updated before it is considered abandoned.
          The abandoned evidence threshold defaults to 30 days if none is
          provided.

        :returns: a list of abandoned evidence files.
        """
        abandoned_evidence = []
        tree = self.repo.head.commit.tree
        for f in tree.traverse():
            if (f.type != 'blob' or f.path.startswith('notifications/')
                    or f.path == 'check_results.json'
                    or f.path.split('/').pop() in AE_EXEMPT):
                continue
            metadata = self.get_evidence_metadata(f.path, dt.utcnow())
            if self._evidence_abandoned(metadata, threshold):
                abandoned_evidence.append(f.path)
        return set(abandoned_evidence)

    def delete_repo_locally(self):
        """Remove the local git repository."""
        try:
            shutil.rmtree(self.local_path)
        except OSError:
            pass

    def get_hash(self):
        """
        Provide the current master revision.

        :returns: the commit SHA or None.
        """
        if self.repo.heads:
            return self.repo.heads.master.commit.hexsha

    def get_latest_commit(self, path, dt=None):
        """
        Provide the most recent commit for the file and date specified.

        :param path: the relative path to the file.
        :param dt: the upper bound commit date.

        :returns: the latest commit containing the file specified.
        """
        commit = None
        options = {'max_count': 1}
        if dt:
            options['until'] = dt.strftime('%Y-%m-%d')
        try:
            commit = next(self.repo.iter_commits(paths=path, **options))
        except StopIteration:
            pass
        return commit

    def get_evidence_metadata(self, evidence_path, evidence_dt=None):
        """
        Provide evidence metadata from package index file.

        :param evidence_path: the evidence relative path.
        :param evidence_dt: the upper bound evidence commit date.

        :returns: the metadata for the evidence specified.
        """
        pkg_index_path = self.get_index_file_by_path(evidence_path)
        if not os.path.exists(pkg_index_path):
            return
        ev_path, ev_name = evidence_path.rsplit('/', 1)
        if evidence_dt:
            pkg_index_path = os.path.join(ev_path, INDEX_FILE)
            commit = self.get_latest_commit(pkg_index_path, evidence_dt)
            if not commit:
                return
            metadata = json.loads(
                commit.tree[pkg_index_path].data_stream.read()
            )
        else:
            metadata = json.loads(open(pkg_index_path).read())
        return metadata.get(
            ev_name,
            self._get_partitioned_evidence_metadata(metadata, ev_name)
        )

    def __enter__(self):
        """
        Initialize the local git repository.

        Either check out the repository from `repo_url` or create a local one.
        """
        self.init()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Handle local locker check-in and remote push if applicable.

        Log an exception if raised, commit the files to the repo and if
        configured push it up to the `repo_url`.
        """
        if exc_type:
            self.logger.error(' '.join([str(exc_type), str(exc_val)]))
        self.checkin()
        if self.repo_url_with_creds:
            self.push()
        return

    def _repo_init(self):
        if os.path.isdir(os.path.join(self.local_path, '.git')):
            self.logger.info(f'Using locker found in {self.local_path}...')
            self.repo = git.Repo(self.local_path)
        else:
            self.logger.info(f'Creating locker in {self.local_path}...')
            self.repo = git.Repo.init(self.local_path)
        self.init_config()

    def _get_partitioned_evidence_metadata(self, metadata, evidence_name):
        try:
            part, ev_name = evidence_name.split('_', 1)
            if part in metadata.get(ev_name, {}).get('partitions', {}).keys():
                return metadata[ev_name]
        except ValueError:
            return

    def _evidence_ttl_expired(self, evidence, last_update):
        expired = False
        last_update_ts = dt.strptime(last_update, '%Y-%m-%dT%H:%M:%S.%f')
        time_diff = dt.utcnow() - last_update_ts
        if time_diff.total_seconds() >= evidence.ttl - self.ttl_tolerance:
            expired = True
        return expired

    def _stale_message(self, message):
        self._handler.setFormatter(logging.Formatter('%(message)s'))
        self.logger.info(message)
        self._handler.setFormatter(
            logging.Formatter('%(levelname)s: %(message)s')
        )

    def _evidence_abandoned(self, metadata=None, threshold=None):
        if metadata is None or metadata.get('last_update') is None:
            return True
        last_update = dt.strptime(
            metadata['last_update'], '%Y-%m-%dT%H:%M:%S.%f'
        )
        time_diff = dt.utcnow() - last_update
        if threshold is None:
            threshold = AE_DEFAULT
        threshold += metadata.get('ttl', 0)
        if time_diff.total_seconds() >= threshold:
            return True
        return False

    def _get_file_content(self, file_path, file_dt=None):
        if not file_dt:
            return open(os.path.join(self.local_path, file_path)).read()
        commit = self.get_latest_commit(file_path, file_dt)
        if not commit:
            raise HistoricalEvidenceNotFoundError(
                f'Evidence {file_path} was not found in the locker '
                f'for {file_dt.strftime("%Y-%m-%d")}'
            )
        return commit.tree[file_path].data_stream.read()

    def _validate_evidence(self, evidence, ignore_ttl):
        if evidence.path in self.forced_evidence:
            raise StaleEvidenceError(
                f'Evidence {evidence.path} is forced stale'
            )

        paths = []
        metadata = self.get_evidence_metadata(evidence.path) or {}
        if getattr(evidence, 'is_partitioned', False):
            for part_hash in metadata['partitions'].keys():
                paths.append(
                    os.path.join(
                        evidence.dir_path, f'{part_hash}_{evidence.name}'
                    )
                )
        else:
            paths.append(evidence.path)
        for path in paths:
            full_path = os.path.join(self.local_path, path)
            if not os.path.exists(full_path):
                raise ValueError(
                    f'Evidence {path} was not found in the locker'
                )
        ttl_expired = self._evidence_ttl_expired(
            evidence, metadata['last_update']
        )
        if not ignore_ttl and ttl_expired:
            raise StaleEvidenceError(f'Evidence {evidence.path} is stale')


def is_index_file(path):
    """Confirm whether the supplied path is to an index file."""
    return os.path.basename(path) == INDEX_FILE



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compliance check automation module."""

import unittest

from compliance.config import get_config

import inflection


class ComplianceCheck(unittest.TestCase):
    """Compliance check automation TestCase class."""

    def __init__(self, *args, **kwargs):
        """Construct and initialize the check test object."""
        super(ComplianceCheck, self).__init__(*args, **kwargs)
        if hasattr(ComplianceCheck, 'config'):
            self.config = ComplianceCheck.config
        else:
            self.config = get_config()

    @property
    def title(self):
        """Check title, normally used in check reports."""
        raise NotImplementedError(
            f'Property title not implemented on {self.__class__.__name__}'
        )

    @property
    def reports(self):
        """List of the check reports."""
        if '_reports' not in self.__dict__.keys():
            self._reports = []
        return self._reports

    @property
    def runbook_url(self):
        """
        Runbook URL associated with the check.

        If runbook links are enabled in the configuration, the value returned
        is based on the runbook base URL and the check's class name.  It is
        assumed that the runbook exists and is in HTML format.
        """
        runbook_enabled = self.config.get('runbooks.enabled')
        runbook_baseurl = self.config.get('runbooks.base_url')
        if runbook_baseurl and runbook_enabled:
            runbook_name = inflection.underscore(self.__class__.__name__)
            return f'{runbook_baseurl}/{runbook_name}.html'

    @property
    def fixed_failure_count(self):
        """Count of failures fixed by the fix routine."""
        if '_fixed_failure_count' not in self.__dict__.keys():
            self._fixed_failure_count = 0
        return self._fixed_failure_count

    @fixed_failure_count.setter
    def fixed_failure_count(self, value):
        self._fixed_failure_count = value

    @property
    def warnings(self):
        """Warnings as a property for each check."""
        if not hasattr(self, '_warnings'):
            self._warnings = {}
        return self._warnings

    @property
    def failures(self):
        """Failures as a property for each check."""
        if not hasattr(self, '_failures'):
            self._failures = {}
        return self._failures

    @property
    def successes(self):
        """Successes as a property for each check."""
        if not hasattr(self, '_successes'):
            self._successes = {}
        return self._successes

    @property
    def evidence_metadata(self):
        """Metadata of all evidence used by each check as a property."""
        if not hasattr(self, '_evidence_metadata'):
            self._evidence_metadata = {}
        return self._evidence_metadata

    def id(self):  # noqa: A003
        """Reset the test id if it has been transplanted and return it."""
        return_id = super(ComplianceCheck, self).id()
        origin = getattr(self, '__origin_module__', None)
        if origin and not return_id.startswith(origin):
            original = '.'.join(
                [self.__origin_module__, self.__class__.__name__]
            )
            transplanted = '.'.join(
                [
                    self.__class__.__module__,
                    getattr(
                        self.__class__,
                        '__qualname__',
                        self.__class__.__name__
                    )
                ]
            )
            return return_id.replace(transplanted, original)
        return return_id

    def total_issues_count(self, results):
        """
        Total number of issues as property for the check.

        :param results: contains all the test results.
        """
        return (
            self.failures_for_check_count(results)
            + self.warnings_for_check_count(results)
        )

    def add_warnings(self, section_key, section_value):
        """
        Add warnings for each check.

        :param: section_key: key for the warnings.
        :param: section_value: all the warnings for that key.
        """
        self.warnings.setdefault(section_key, [])
        if isinstance(section_value, (list, set)):
            self.warnings[section_key].extend(section_value)
        else:
            self.warnings[section_key].append(section_value)

    def warnings_for_check(self, results):
        """
        Provide warnings associated with all tests in the check.

        :param results: contains all the test results.
        """
        w = {}
        for t in self._get_all_test_objs(results):
            for k, v in t.warnings.items():
                w.setdefault(k, []).extend(v)
        return w

    def warnings_for_check_count(self, results):
        """
        Count failures for all tests in the check.

        :param results: contains all the test results.
        """
        return sum(
            t.warnings_count() for t in self._get_all_test_objs(results)
        )

    def warnings_count(self):
        """Count warnings for a specific test."""
        return sum(len(v) for _, v in self.warnings.items())

    def add_failures(self, section_key, section_value):
        """
        Add failures for each check.

        :param: section_key: key for the failures.
        :param: section_value: all the failures for that key.
        """
        self.failures.setdefault(section_key, [])
        if isinstance(section_value, (list, set)):
            self.failures[section_key].extend(section_value)
        else:
            self.failures[section_key].append(section_value)

    def failures_for_check(self, results):
        """
        Provide failures associated with all tests in the check.

        :param results: contains all the test results.
        """
        f = {}
        for t in self._get_all_test_objs(results):
            for k, v in t.failures.items():
                f.setdefault(k, []).extend(v)
        return f

    def failures_for_check_count(self, results):
        """
        Count failures for all tests in check.

        :param results: contains all the test results.
        """
        return sum(
            t.failures_count() for t in self._get_all_test_objs(results)
        )

    def failures_count(self):
        """Count failures for a specific test."""
        return sum(len(v) for _, v in self.failures.items())

    def add_successes(self, section_key, section_value):
        """
        Add successes for each check.

        :param: section_key: key for the successes.
        :param: section_value: all the successes for that key.
        """
        self.successes.setdefault(section_key, [])
        if isinstance(section_value, (list, set)):
            self.successes[section_key].extend(section_value)
        else:
            self.successes[section_key].append(section_value)

    def successes_for_check(self, results):
        """
        Provide successes associated with all tests in the check.

        :param results: contains all the test results.
        """
        s = {}
        for t in self._get_all_test_objs(results):
            for k, v in t.successes.items():
                s.setdefault(k, []).extend(v)
        return s

    def successes_for_check_count(self, results):
        """
        Count successes for all tests in check.

        :param results: contains all the test results.
        """
        return sum(
            t.successes_count() for t in self._get_all_test_objs(results)
        )

    def successes_count(self):
        """Count successes for a specific test."""
        return sum(len(v) for _, v in self.successes.items())

    def run(self, result=None):
        """
        Override the unittest.TestCase.run method.

        Decorates ``test_`` methods by adding common functionality
        (e.g. failures count check).
        """

        def wrapper(method):

            def check_failures():
                method()
                self.assertEquals(self.failures_count(), 0)

            return check_failures

        test_method = getattr(self, self._testMethodName)
        setattr(self, self._testMethodName, wrapper(test_method))
        super(ComplianceCheck, self).run(result)

    def add_issue_if_diff(self, actual, expected, msg, as_warning=False):
        """
        Add failures/warnings if differences exist between actual and expected.

        :param actual: the "actual" set items.
        :param expected: the "expected" set items.
        :param msg: the section key used if differences are found.
        :param as_warning: warn if True, fail if False.
        """
        diff = set(actual) - set(expected)
        if not diff:
            return
        if as_warning:
            self.add_warnings(msg, sorted(diff))
        else:
            self.add_failures(msg, sorted(diff))

    def get_historical_evidence(self, evidence_path, evidence_dt):
        """
        Retrieve historical evidence from the locker and track as metadata.

        :param evidence_path: the evidence path.
        :param evidence_dt: the evidence date.
        """
        evidence = self.locker.get_evidence(evidence_path, True, evidence_dt)
        self.add_evidence_metadata(evidence_path, evidence_dt)
        return evidence

    def add_evidence_metadata(self, evidence_path, evidence_dt=None):
        """
        Add evidence metadata to the evidences property of each check.

        :param evidence_path: the evidence path.
        :param evidence_dt: the evidence date.
        """
        metadata = self.locker.get_evidence_metadata(
            evidence_path, evidence_dt
        )
        metadata['path'] = evidence_path
        if metadata.get('partitions'):
            path, file_name = evidence_path.rsplit('/', 1)
            partitions = {}
            for part_hash, part_key in metadata['partitions'].items():
                partitions[part_hash] = {
                    'key': part_key,
                    'commit_sha': self.locker.get_latest_commit(
                        f'{path}/{part_hash}_{file_name}', evidence_dt
                    ).hexsha
                }
            metadata['partitions'] = partitions
        else:
            metadata['commit_sha'] = self.locker.get_latest_commit(
                evidence_path, evidence_dt
            ).hexsha
        metadata.pop('tombstones', None)
        evidence_meta_key = (evidence_path, metadata['last_update'])
        self.evidence_metadata[evidence_meta_key] = metadata

    def evidences_for_check(self, results):
        """
        Provide evidences used by all tests in the check.

        :param results: contains all the test results.
        """
        evidences = {}
        for test_obj in self._get_all_test_objs(results):
            evidences.update(test_obj.evidence_metadata)
        return list(evidences.values())

    def _get_all_test_objs(self, results):
        """
        Provide all completed tests from the test results.

        All completed tests that are associated with this
        check class are returned.
        """
        return [
            info['test'].test
            for info in results.values()
            if info['test'].test.__class__ == self.__class__
        ]



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compliance credentials configuration."""

import logging
import os
from collections import OrderedDict, namedtuple
from configparser import RawConfigParser

# used to differentiate a user passing None vs
# not passing a value for an optional argument
_sentinel = object()

logger = logging.getLogger('compliance.utils.credentials')


class Config():
    """Handle credentials configuration."""

    def __init__(self, cfg_file='~/.credentials'):
        """
        Create an instance of a dictionary-like configuration object.

        :param cfg_file: The path to the RawConfigParser compatible config file
        """
        self._cfg = RawConfigParser()
        self._cfg.read(os.path.expanduser(cfg_file))
        self._cfg_file = cfg_file

    def __getitem__(self, section):
        """
        Get the named tuple representing the configuration held at `section`.

        Build a named tuple representing the configuration at `section`. If a
        config file does not have an option for the section ignore it.
        Resulting in an AttributeError if accessed later in the code.

        :param section: the section to retrieve
        """

        def _getattr_wrapper(t, attr):
            """
            Replace the standard __getattr__ functionality.

            In the case when a section and/or attribute is not set in the
            config file, the error shown will be more helpful.
            """
            try:
                return t.__getattribute__(attr)
            except AttributeError as exc:
                exc.args = (
                    (
                        f'Unable to locate attribute "{attr}" '
                        f'in section "{type(t).__name__}" '
                        f'at config file "{self._cfg_file}"'
                    ),
                )
                raise exc

        env_vars = [
            k for k in os.environ.keys()
            if k.startswith(f'{section.upper()}_')
        ]
        env_keys = [
            k.split(section.upper())[1].lstrip('_').lower() for k in env_vars
        ]
        env_values = [os.environ[e] for e in env_vars]
        if env_vars:
            logger.debug(
                f'Loading credentials from ENV vars: {", ".join(env_vars)}'
            )
        params = []
        if self._cfg.has_section(section):
            params = self._cfg.options(section)
        values = [self._cfg.get(section, x) for x in params]

        d = OrderedDict(zip(params, values))

        if env_vars:
            d.update(zip(env_keys, env_values))
        t = namedtuple(section, ' '.join(list(d.keys())))
        t.__getattr__ = _getattr_wrapper
        return t(*list(d.values()))

    def get(self, section, key=None, account=None, default=_sentinel):
        """
        Retrieve sections and keys by account.

        ``section``: the section from which to retrieve keys.
        ``key``: the key in the section whose value you want to retrieve. if
            not specified, returns the whole section as a dictionary.
        ``account``: if provided, fetches the value for the specific account.
            assumes the account is prefixed to the key and separated by _.
        ``default``: if provided, returns this value if a value cannot be
            found; otherwise raises an exception.
        """
        if key is None:
            return self[section]
        if account:
            key = '_'.join([account, key])
        if default == _sentinel:
            return getattr(self[section], key)
        else:
            return getattr(self[section], key, default)



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compliance notification management automation module."""

import copy
import json
import logging
import os
import sys
import time
from datetime import datetime
from urllib.parse import urlparse

from compliance.config import get_config
from compliance.utils.services import pagerduty
from compliance.utils.services.github import Github
from compliance.utils.test import parse_test_id

from ibm_cloud_sdk_core.authenticators import IAMAuthenticator

from ibm_security_advisor_findings_api_sdk import ApiException, FindingsApiV1

import requests


class _BaseNotifier(object):
    """
    Base notifier class.

    It shouldn't be used outside of this module.
    """

    def __init__(self, results, controls, push_error):
        self._results = results
        self._controls = controls
        self._push_error = push_error
        self.logger = logging.getLogger(name='compliance.notifier')
        self._handler = logging.StreamHandler()
        self._handler.setFormatter(
            logging.Formatter('%(levelname)s: %(message)s')
        )
        self.logger.handlers.clear()
        self.logger.addHandler(self._handler)
        self.logger.setLevel(logging.INFO)

    @property
    def messages(self):
        """
        Check test messages.

        A generator of list of tuples containing the following structure::

          ([str] test_id, [dict] test_descriptor, [dict] message)
        """
        for test_id, test_desc in self._results.items():
            test_obj = test_desc['test'].test
            method_name = parse_test_id(test_id)['method']

            msg_method = 'get_notification_message'
            if len(test_obj.tests) > 1:
                candidate = method_name.replace('test_', 'msg_', 1)
                if hasattr(test_obj, candidate):
                    msg_method = candidate

            # set body to None if the notification function hasn't been
            # defined or if it returns None.
            # use a predefined error message for error status.
            # otherwise get the results of the notification function.
            # note that passed tests get their notifications called in order to
            # deduce things like subtitle, but their notifications are not
            # displayed.
            if not hasattr(test_obj, msg_method):
                msg = None
                body = None
            elif test_desc['status'] == 'error':
                msg = None
                body = f'Check {test_id} failed to execute'
            elif len(test_obj.tests) > 1 and not msg_method.startswith('msg_'):
                msg = getattr(test_obj, msg_method)(method_name)
                body = msg and 'body' in msg and msg['body'] or None
            else:
                msg = getattr(test_obj, msg_method)()
                body = msg and 'body' in msg and msg['body'] or None

            title = test_obj.title
            if msg and 'subtitle' in msg and msg['subtitle']:
                title += f' - {msg["subtitle"]}'

            failure_count = 0
            if msg and test_obj.failures:
                failure_count = test_obj.failures_count()

            warning_count = 0
            if msg and test_obj.warnings:
                warning_count = test_obj.warnings_count()

            msg = {
                'title': title,
                'body': body,
                'failure_count': failure_count,
                'warning_count': warning_count
            }
            yield test_id, test_desc, msg

    def _messages_by_accreditations(self):
        retval = {}
        for test_id, test_desc, msg in self.messages:
            test_class = parse_test_id(test_id)['class_path']
            accreditations = self._controls.get_accreditations(test_class)
            for a in accreditations:
                messages = retval.get(a, [])
                messages.append((test_id, test_desc, msg))
                retval[a] = messages
        return retval

    def _split_by_status(self, messages):
        passed_tests = []
        failed_tests = []
        warned_tests = []
        errored_tests = []
        sorted_msgs = sorted(messages, key=lambda x: x[2]['title'])
        for test_id, test_desc, msg in sorted_msgs:
            if test_desc['status'] == 'pass':
                passed_tests.append((test_id, test_desc, msg))
            elif test_desc['status'] == 'error':
                errored_tests.append((test_id, test_desc, msg))
            elif test_desc['status'] == 'warn':
                warned_tests.append((test_id, test_desc, msg))
            else:
                failed_tests.append((test_id, test_desc, msg))
        return passed_tests, failed_tests, warned_tests, errored_tests

    def _get_check_names(self, checks, include_path=False):
        if include_path:
            return [(msg['title'], path) for path, _, msg in checks]
        return [msg['title'] for _, _, msg in checks]

    def _get_report_links(self, test_desc, link_format=None):
        if not link_format:
            link_format = '<{url}|{name}>'

        if test_desc['status'] == 'error':
            return []

        test_obj = test_desc['test'].test

        return [
            link_format.format(
                url=test_obj.locker.get_remote_location(report.path),
                name=report.name
            ) for report in test_obj.reports
        ]

    def _get_summary_and_body(
        self,
        test_desc,
        msg,
        include_title=True,
        summary_format=None,
        link_format=None
    ):
        link_format = link_format or '<{url}|{name}>'
        if not summary_format:
            summary_format = ''
            if include_title:
                summary_format += '{title} - '
            summary_format += (
                '{status} ({issues}) '
                'Reports: {reports} {runbook}'
            )

        test_obj = test_desc['test'].test

        issues_list = []
        if msg['failure_count'] > 0:
            issues_list.append(f'{msg["failure_count"]} failures')
        if msg['warning_count'] > 0:
            issues_list.append(f'{msg["warning_count"]} warnings')
        if test_obj.fixed_failure_count > 0:
            issues_list.append(f'{test_obj.fixed_failure_count} fixed')

        issues = ', '.join(issues_list)
        report_links = self._get_report_links(test_desc, link_format)
        if not report_links:
            report_links.append('(none)')

        runbook_conditional = ''
        if test_obj.runbook_url:
            runbook_conditional = '| ' + link_format.format(
                url=test_obj.runbook_url, name='Run Book'
            )

        summary_line = summary_format.format(
            title=msg['title'],
            status=test_desc['status'].upper(),
            issues=issues,
            reports=', '.join(report_links),
            runbook=runbook_conditional
        )

        body = msg['body'] and f'\n{msg["body"]}' or ''

        return summary_line, body


class _BaseMDNotifier(_BaseNotifier):
    """
    Base markdown notifier class.

    It shouldn't be used outside of this module.
    """

    def __init__(self, results, controls, push_error):
        super(_BaseMDNotifier, self).__init__(results, controls, push_error)

    def _generate_accred_content(self, accred, results, skip_title=False):
        md_content = []
        if not skip_title:
            now = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S UTC')
            md_content.append(f'# CHECK RESULTS: {now}')
        md_content.append(
            f'\n## Notification for {accred.upper()} accreditation\n'
        )
        if self._push_error:
            md_content.append('### All Checks (Errored)\n')
            md_content.append(
                '   - Evidence/Results failed to push to remote locker.  '
                'See execution log for details.'
            )
        else:
            for heading in ['Passed Checks', 'Errored Checks']:
                md_content.append(f'### {heading}\n')
                checks = self._get_check_names(
                    results[heading[:-9].lower()], include_path=True
                )
                if checks:
                    check_title = None
                    for check in checks:
                        if check[0] != check_title:
                            md_content.append(f'- **{check[0]}**')
                            check_title = check[0]
                        if heading == 'Errored Checks':
                            md_content.append(
                                f'   - {check[1]} failed to execute'
                            )
                else:
                    md_content.append(f'- **No {heading.lower()}**')
            md_content.append('### Failures/Warnings\n')
            fail_and_warn = results['fail'] + results['warn']
            if fail_and_warn:
                summary_format = [
                    '- **{title}**',
                    '   - **{status}** | {reports} {runbook}',
                    '   - {issues}'
                ]
                rpt_link_format = '[{name}]({url})'
                for _, test_desc, msg in fail_and_warn:
                    summary, addl_content = self._get_summary_and_body(
                        test_desc,
                        msg,
                        summary_format='\n'.join(summary_format),
                        link_format=rpt_link_format
                    )
                    md_content.append(summary)
                    if addl_content:
                        for line in addl_content.strip().split('\n'):
                            md_content.append(f'   - _{line}_')
            else:
                md_content.append('- **No failures or warnings**')
        return md_content


class FDNotifier(_BaseNotifier):
    """
    File descriptor notifier class.

    Notifications are written to the file descriptor specified.
    Defaults to STDOUT.
    """

    def __init__(self, results, controls, fd=sys.stdout, push_error=False):
        """
        Construct and initialize the file descriptor notifier object.

        :param results: dictionary generated by
          :py:class:`compliance.runners.CheckMode` at the end of the execution.
        :param controls: the control descriptor that manages accreditations.
        :param fd: a file descriptor where to write the notifications on.
          Defaults to STDOUT.
        """
        super(FDNotifier, self).__init__(results, controls, push_error)
        self.fd = fd

    def notify(self):
        """Write notifications into the file descriptor."""
        self.logger.info('Running the STDOUT notifier...')
        self.fd.write('\n-- NOTIFICATIONS --\n\n')
        if not self._results:
            self.fd.write('No results\n')
        elif self._push_error:
            self.fd.write(
                'All accreditation checks:  '
                'Evidence/Results failed to push to remote locker.\n'
            )
        else:
            accreds = []
            messages = list(self._messages_by_accreditations().items())
            messages.sort(key=lambda x: x[0])
            for accreditation, msgs in messages:
                if not msgs:
                    continue
                passed, failed, warned, errored = self._split_by_status(msgs)
                accreds.append(
                    {
                        'name': accreditation,
                        'passed': passed,
                        'failed': failed,
                        'warned': warned,
                        'errored': errored
                    }
                )
            for accred in accreds:
                self.fd.write(
                    f'Notifications for {accred["name"].upper()} '
                    'accreditation\n\n'
                )
                passed_msg = ', '.join(
                    self._get_check_names(accred['passed'])
                ) or '(none)'
                accred_msgs = [f'PASSED checks: {passed_msg}']
                for msg_type in ['errored', 'warned', 'failed']:
                    accred_msgs.append(
                        '\n\n'.join(
                            [
                                ''.join(
                                    self._get_summary_and_body(test_desc, msg)
                                ) for (_, test_desc, msg) in accred[msg_type]
                            ]
                        )
                    )
                self.fd.write('\n\n'.join(accred_msgs) + '\n\n')
        self.fd.flush()


class LockerNotifier(_BaseMDNotifier):
    """
    Evidence Locker notifier class.

    Notifications are written to the evidence locker.

    :param results: dictionary generated by
      :py:class:`compliance.runners.CheckMode` at the end of the execution.
    :param controls: dictionary of checks and the accreditations and controls
      that they belong to.
    """

    def __init__(self, results, controls, locker, push_error=False):
        """
        Construct and initialize the evidence locker notifier object.

        :param results: dictionary generated by
          :py:class:`compliance.runners.CheckMode` at the end of the execution.
        :param controls: the control descriptor that manages accreditations.
        :param locker: the evidence locker object.
        """
        super(LockerNotifier, self).__init__(results, controls, push_error)
        self.locker = locker

    def notify(self):
        """Write notifications into the evidence locker."""
        if not self._results:
            self.logger.error('No results.  Locker notifier not triggered.')
            return
        if self._push_error:
            self.logger.error(
                'Remote locker push failed.  Locker notifier not triggered.'
            )
            return
        self.logger.info('Running the Locker notifier...')
        messages = list(self._messages_by_accreditations().items())
        messages.sort(key=lambda x: x[0])
        now = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S UTC')
        md_content = [f'# CHECK RESULTS: {now}']
        for accreditation, results in messages:
            passed, failed, warned, errored = self._split_by_status(results)
            results_by_status = {
                'pass': passed,
                'fail': failed,
                'warn': warned,
                'error': errored
            }
            md_content.extend(
                self._generate_accred_content(
                    accreditation, results_by_status, skip_title=True
                )
            )
        folder = 'notifications'
        filename = 'alerts_summary.md'
        self.locker.add_content_to_locker(
            '\n'.join(md_content), folder, filename
        )
        self.locker.checkin(
            'Locker notification sent at local time '
            f'{time.ctime(time.time())}\n\n{os.path.join(folder, filename)}'
        )
        self.locker.push()


class GHIssuesNotifier(_BaseMDNotifier):
    """
    Github notifier class.

    Notifications are sent to Github as repository issues.  This
    notifier is configurable via :class:`compliance.config.ComplianceConfig`.
    """

    def __init__(self, results, controls, push_error=False):
        """
        Construct and initialize the Github notifier object.

        :param results: dictionary generated by
          :py:class:`compliance.runners.CheckMode` at the end of the execution.
        :param controls: the control descriptor that manages accreditations.
        """
        super(GHIssuesNotifier, self).__init__(results, controls, push_error)

        self._config = get_config().get('notify.gh_issues')
        if not self._config:
            # Ensure that legacy ghe_issues config still works
            self._config = get_config().get('notify.ghe_issues', {})
        # Using the locker repo url to define the base url.  The expectation
        # is that the Github issues repository will share the base url.
        parsed_locker_url = urlparse(get_config().get('locker.repo_url'))
        self._github = Github(
            get_config().creds,
            f'{parsed_locker_url.scheme}://{parsed_locker_url.hostname}'
        )

    def notify(self):
        """Send notifications to Github as repository issues."""
        self.logger.info('Running the Github Issues notifier...')
        if not self._config:
            self.logger.warning('Using Github Issues notifier without config')

        messages = list(self._messages_by_accreditations().items())
        messages.sort(key=lambda x: x[0])
        for accreditation, results in messages:
            if accreditation not in self._config:
                continue
            passed, failed, warned, errored = self._split_by_status(results)
            results_by_status = {
                'pass': passed,
                'fail': failed,
                'warn': warned,
                'error': errored
            }
            if self._config[accreditation].get('summary_issue'):
                self._notify_by_summary_issue(accreditation, results_by_status)
            elif self._push_error:
                self.logger.error(
                    'Remote locker push failed.  '
                    'Github Issues notifier not triggered.'
                )
            else:
                self._notify_by_check_issues(accreditation, results_by_status)

    def _notify_by_summary_issue(self, accred, results):
        issue = [self._generate_summary_issue(accred, results)]
        repos = self._config[accred].get('repo', [])
        for repo in repos:
            owner, repository = repo.split('/')
            issue_urls = self._process_new_alerts(
                owner,
                repository,
                issue,
                self._config[accred]['summary_issue'].get('message')
            )
            self._assign_projects(issue_urls, repo, accred)

    def _generate_summary_issue(self, accred, results):
        summary_config = self._config[accred]['summary_issue']
        title = summary_config['title']
        labels = summary_config.get('labels', [])
        assignees = summary_config.get('assignees', [])
        frequency = summary_config.get('frequency')
        rotation = summary_config.get('rotation')
        rotation_index = None
        now = datetime.utcnow()
        if frequency == 'day':
            today = now.strftime('%Y-%m-%d')
            title = f'{today} - {title}'
            labels.extend([frequency, today])
            rotation_index = now.timetuple().tm_yday
        elif frequency == 'week':
            year, week, _ = now.isocalendar()
            title = f'{year}, {week}W - {title}'
            labels.extend([frequency, str(year), f'{week}W'])
            rotation_index = week
        elif frequency == 'month':
            year = now.strftime('%Y')
            month = now.strftime('%mM')
            title = f'{year}, {month} - {title}'
            labels.extend([frequency, year, month])
            rotation_index = int(month[:-1])
        elif frequency == 'year':
            year = now.strftime('%Y')
            title = f'{year} - {title}'
            labels.extend([frequency, year])
            rotation_index = int(year)
        if rotation and rotation_index:
            assignees = rotation[divmod(rotation_index, len(rotation))[1]]
        issue = {'title': title, 'labels': labels, 'assignees': assignees}
        issue['body'] = '\n'.join(
            self._generate_accred_content(accred, results)
        )
        return issue

    def _notify_by_check_issues(self, accred, results):
        issues = []
        statuses = self._config[accred].get('status', ['fail'])
        repos = self._config[accred].get('repo', [])
        for status, result in results.items():
            if status in statuses:
                issues += self._generate_issues(accred, result)
        for repo in repos:
            owner, repository = repo.split('/')
            issue_urls = self._process_new_alerts(owner, repository, issues)
            self._assign_projects(issue_urls, repo, accred)
        if 'pass' not in statuses:
            for repo in repos:
                owner, repository = repo.split('/')
                issues = self._generate_issues(accred, results['pass'])
                issue_urls = self._process_old_alerts(
                    owner, repository, issues
                )
                self._assign_projects(issue_urls, repo, accred)

    def _generate_issues(self, accred, results):
        issues = []
        if not results:
            return issues
        for check_path, result, message in results:
            now = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S UTC')
            body = [f'## Compliance check alert - {now}']
            body.append(f'- Check: {check_path}')
            test_obj = result['test'].test
            check_name = check_path.rsplit('.', 1).pop()
            doc = getattr(test_obj.__class__, check_name).__doc__
            if doc:
                doc = doc.strip()
                newline = doc.find('\n')
                if newline > -1:
                    doc = doc[:newline]
                body.append(f'- Description: {doc}')
            body.append(f'- Accreditation: {accred}')
            status = ''.join(
                self._get_summary_and_body(
                    result,
                    message,
                    include_title=False,
                    summary_format='{status} ({issues})',
                    link_format='[{name}]({url})'
                )
            )
            body.append(f'- Run Status: **{status}**')
            run_dttm = datetime.fromtimestamp(result['timestamp'])
            body.append(f'- Run Date/Time: {run_dttm}')
            report_links = self._get_report_links(
                result, link_format='[{name}]({url})'
            )
            if report_links:
                body.append(f'- Reports: {", ".join(report_links)}')

            issue = {
                'title': message['title'],
                'body': '\n'.join(body),
                'labels': [
                    f'accreditation: {accred}',
                    f'run status: {result["status"]}'
                ]
            }
            issues.append(issue)

        return issues

    def _process_new_alerts(self, owner, repository, issues, message=None):
        issue_urls = {}
        for issue in issues:
            gh_issue = self._find_gh_issue(
                '/'.join([owner, repository]), issue['title']
            )
            if gh_issue is None:
                body = issue['body']
                if message:
                    joined_msg = '\n'.join(message)
                    body = f'{joined_msg}\n\n{issue["body"]}'
                gh_issue = self._github.add_issue(
                    owner,
                    repository,
                    issue['title'],
                    body,
                    labels=issue['labels'],
                    assignees=issue.get('assignees', [])
                )
            else:
                self._update_issue_labels(
                    owner, repository, gh_issue, issue['labels']
                )
                self._github.add_issue_comment(
                    owner, repository, gh_issue['number'], issue['body']
                )
            issue_urls[gh_issue['id']] = gh_issue['url']
        return issue_urls

    def _process_old_alerts(self, owner, repository, issues):
        issue_urls = {}
        for issue in issues:
            gh_issue = self._find_gh_issue(
                '/'.join([owner, repository]), issue['title']
            )
            if gh_issue:
                self._update_issue_labels(
                    owner, repository, gh_issue, issue['labels']
                )
                self._github.add_issue_comment(
                    owner, repository, gh_issue['number'], issue['body']
                )
                issue_urls[gh_issue['id']] = gh_issue['url']
        return issue_urls

    def _find_gh_issue(self, repo, title):
        gh_issues = self._github.search_issues(
            f'{title} type:issue in:title is:open repo:{repo}'
        )
        found = None
        for issue in gh_issues:
            if issue['title'] == title:
                found = issue
                break
        return found

    def _update_issue_labels(self, owner, repository, issue, labels):
        current_labels = [label['name'] for label in issue['labels']]
        new_labels = list(set(labels) - set(current_labels))
        if new_labels:
            current_labels = [
                label for label in current_labels
                if not label.startswith('run status: ')
            ]
            self._github.patch_issue(
                owner,
                repository,
                issue['number'],
                labels=current_labels + new_labels
            )

    def _assign_projects(self, issues, repo, accred):
        config_projects = self._config[accred].get('project')
        if not config_projects:
            return
        all_projects = {
            p['name']: p['id']
            for p in self._github.get_all_projects(repo)
        }
        for project, column in config_projects.items():
            if project not in all_projects.keys():
                self.logger.warning(f'Project {project} not found in {repo}')
                continue
            columns = {
                c['name']: c['id']
                for c in self._github.get_columns(all_projects[project])
            }
            if column not in columns.keys():
                self.logger.warning(
                    f'Column {column} not found '
                    f'in {project} project, {repo} repo'
                )
                continue
            card_lists = self._github.get_all_cards(columns.values()).values()
            issue_urls = [
                c.get('content_url') for cl in card_lists for c in cl
            ]
            for issue_id, issue_url in issues.items():
                if issue_url in issue_urls:
                    continue
                self._github.add_card(columns[column], issue=issue_id)


class SlackNotifier(_BaseNotifier):
    """
    Slack notifier class.

    Notifications are sent to Slack channel(s).  This notifier is
    configurable via :class:`compliance.config.ComplianceConfig`.
    """

    MESSAGE_COLORS = {
        'pass': '#00D000',
        'fail': '#D00000',
        'error': '#9932CC',
        'warn': '#FFD300'
    }

    def __init__(self, results, controls, push_error=False):
        """
        Construct and initialize the Slack notifier object.

        :param results: dictionary generated by
          :py:class:`compliance.runners.CheckMode` at the end of the execution.
        :param controls: the control descriptor that manages accreditations.
        """
        super(SlackNotifier, self).__init__(results, controls, push_error)
        self._creds = get_config().creds
        self._config = get_config().get('notify.slack', {})

    def notify(self):
        """Send notifications to Slack channel(s)."""
        self.logger.info('Running the Slack notifier...')
        if not self._config:
            self.logger.warning('Using Slack notifier without config')

        messages = list(self._messages_by_accreditations().items())
        messages.sort(key=lambda x: x[0])
        for accreditation, desc in messages:
            if accreditation not in self._config:
                continue
            channels = []
            mode = 'normal'
            if isinstance(self._config[accreditation], list):
                channels = self._config[accreditation]
            elif isinstance(self._config[accreditation], dict):
                channels = self._config[accreditation].get('channels', [])
                rotation = self._config[accreditation].get('rotation')
                mode = self._config[accreditation].get('mode', 'normal')
                if rotation and isinstance(rotation, list):
                    iso_week = datetime.utcnow().isocalendar()[1]
                    on_duty = rotation[divmod(iso_week, len(rotation))[1]]
                    if isinstance(on_duty, dict):
                        channels.append(on_duty.get('id'))
                    else:
                        channels.append(on_duty)
            msg = self._generate_message(accreditation, desc, mode=mode)
            self._send_message(msg, channels)

    def _generate_message(self, accreditation, test_descs, mode='normal'):
        if not test_descs:
            return {}
        text = (
            f'Notification for {accreditation.upper()} accreditation '
            f'at {datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S UTC")}'
        )
        message = {
            'username': 'Compliance Alerts',
            'icon_emoji': ':robot_face:',
            'text': text
        }
        modes = {
            'normal': self._generate_normal_attachments,
            'compact': self._generate_compact_attachments
        }
        if mode in modes.keys():
            if self._push_error:
                message['attachments'] = self._generate_push_error_attachment()
            else:
                message['attachments'] = modes[mode](test_descs)
        else:
            raise ValueError(f'Unknown Slack message mode: {mode}')
        return message

    def _generate_push_error_attachment(self):
        return [
            {
                'title': 'ALL checks',
                'text': (
                    'Evidence/Results failed to push to remote locker.  '
                    'See execution log for details.'
                ),
                'mrkdwn_in': ['text', 'pretext'],
                'color': SlackNotifier.MESSAGE_COLORS['error']
            }
        ]

    def _generate_normal_attachments(self, test_descs):
        retval = []
        passed, failed, warned, errored = self._split_by_status(test_descs)

        # first list each error, failure, and warning
        for _, test_desc, msg in (errored + failed + warned):
            text = ''.join(
                self._get_summary_and_body(
                    test_desc, msg, include_title=False
                )
            )

            attachment = {
                'title': msg['title'],
                'text': text,
                'mrkdwn_in': ['text', 'pretext'],
                'color': SlackNotifier.MESSAGE_COLORS[test_desc['status']]
            }
            retval.append(attachment)

        # then have a list of passed checks
        passed_titles = self._get_check_names(passed)
        if not passed_titles:
            passed_titles.append('(none)')
        retval.append(
            {
                'title': 'PASSED checks',
                'text': ', '.join(passed_titles),
                'mrkdwn_in': ['text', 'pretext'],
                'color': SlackNotifier.MESSAGE_COLORS['pass']
            }
        )

        return retval

    def _generate_compact_attachments(self, test_descs):
        retval = []
        passed, failed, warned, errored = self._split_by_status(test_descs)

        # passed tests
        if passed:
            retval.append(
                {
                    'title': f'PASS: {len(passed)} checks',
                    'text': '',
                    'mrkdwn_in': ['text', 'pretext'],
                    'color': SlackNotifier.MESSAGE_COLORS['pass']
                }
            )

        # warned and failed tests
        collections = [('warn', warned), ('fail', failed)]
        fmt = '{title} - {reports} {runbook} - ({issues})'
        for status, c in collections:
            text = ''
            for _, test_desc, msg in c:
                summary, body = self._get_summary_and_body(
                    test_desc,
                    msg,
                    include_title=False,
                    summary_format=fmt
                )
                text += '* ' + summary
                if body:
                    text += ' - ' + '; '.join(
                        body.strip().replace('Failures:', '').split('\n')
                    )
                text += '\n'
            if text:
                attachment = {
                    'title': f'{status.upper()}: {len(c)} checks',
                    'text': text.strip(),
                    'mrkdwn_in': ['text', 'pretext'],
                    'color': SlackNotifier.MESSAGE_COLORS[status]
                }
                retval.append(attachment)

        # errors tests
        if errored:
            retval.append(
                {
                    'title': f'ERRORS: {len(errored)} checks',
                    'text': ', '.join(set(self._get_check_names(errored))),
                    'mrkdwn_in': ['text', 'pretext'],
                    'color': SlackNotifier.MESSAGE_COLORS['error']
                }
            )
        return retval

    def _send_message(self, message, channels):
        msg = copy.deepcopy(message)
        for c in channels:
            msg['channel'] = c
            headers = {}
            url = (
                getattr(self._creds['cloobot'], 'webhook', None)
                or getattr(self._creds['slack'], 'webhook', None)
            )
            if not url:
                token = (
                    getattr(self._creds['cloobot'], 'token', None)
                    or getattr(self._creds['slack'], 'token', None)
                )
                if token is None:
                    raise RuntimeError(
                        'Unable to get a Slack webhook or token from '
                        'credentials file'
                    )
                url = 'https://slack.com/api/chat.postMessage'
                headers['Authorization'] = 'Bearer ' + token
                headers['Content-type'] = 'application/json'
            retries = self._config.get('retries', 3)
            retry = 0
            while retry < retries:
                response = requests.post(
                    url, headers=headers, data=json.dumps(msg)
                )
                if response.status_code == 429:
                    time.sleep(
                        int(response.headers.get('Retry-After', retry)) + 1
                    )
                    retry += 1
                else:
                    response.raise_for_status()
                    break


class PagerDutyNotifier(_BaseNotifier):
    """PagerDuty notifier class."""

    def __init__(self, results, controls, push_error=False):
        """
        Construct and initialize the PagerDuty notifier object.

        :param results: dictionary generated by
          :py:class:`compliance.runners.CheckMode` at the end of the execution.
        :param controls: the control descriptor that manages accreditations.
        """
        super(PagerDutyNotifier, self).__init__(results, controls, push_error)
        self._creds = get_config().creds
        self._config = get_config().get('notify.pagerduty', {})

    def notify(self):
        """Send notifications as PagerDuty alerts."""
        if self._push_error:
            self.logger.error(
                'Remote locker push failed.  PagerDuty notifier not triggered.'
            )
            return
        self.logger.info('Running the PagerDuty notifier...')
        if not self._config:
            self.logger.warning('Using PagerDuty notifier without config')

        # get all checks by accreditation
        messages = list(self._messages_by_accreditations().items())
        messages.sort(key=lambda x: x[0])
        for accreditation, desc in messages:
            if accreditation not in self._config or not desc:
                continue
            # determine if we should page for all failing checks in this
            # accreditation, or only a subset defined in the config.
            # The config can either be a string of the PD service ID,
            # or a dictionary containing additional config details, including
            # the optional list of checks.
            all_checks, pd_checks = True, []
            if (isinstance(self._config[accreditation], dict)
                    and 'checks' in self._config[accreditation]):
                all_checks = False
                pd_checks = self._config[accreditation]['checks'] or []

            # get all current PD alerts
            alerts = self._get_alerts(accreditation)

            # fail/error will trigger alerts, and pass/warn will clear them
            passed, failed, warned, errored = self._split_by_status(desc)

            for test_id, test_desc, msg in (failed + errored):
                # is the check in scope?
                if not all_checks and test_id not in pd_checks:
                    continue

                summary, body = self._get_summary_and_body(
                    test_desc,
                    msg,
                    summary_format='{status} ({issues})'
                )
                details = summary + body

                # don't trigger an alert if it's already in PD, unless its
                # details have changed
                existing_alert = next(
                    (a for a in alerts if a['test'] == test_id), None
                )
                if existing_alert and existing_alert['details'] == details:
                    continue

                if existing_alert:
                    self._update_alert(
                        test_id, test_desc, msg, accreditation, details
                    )
                else:
                    self._trigger_alert(
                        test_id, test_desc, msg, accreditation, details
                    )

            for test_id, test_desc, msg in (passed + warned):
                # don't resolve an alert if it isn't in PD
                existing_alert = next(
                    (a for a in alerts if a['test'] == test_id), None
                )
                if not existing_alert:
                    continue

                self._resolve_alert(test_id, test_desc, msg, accreditation)

    def _get_alerts(self, accreditation):
        conf = self._config[accreditation]
        pd_service = conf['service_id'] if isinstance(conf, dict) else conf

        # Get all triggered alerts
        # NOTE: statuses[]=triggered is supposed to work, but doesn't
        # so we have to filter the response. leaving it in the query,
        # as it will be a bit more efficient if they fix the issue.
        pages = pagerduty.get(
            'alerts',
            params={
                'service_ids[]': pd_service,
                'statuses[]': 'triggered',
                'time_zone': 'UTC'
            },
            creds=self._creds
        )
        alerts = [
            {
                'test': alert['alert_key'],
                'details': alert.get('body', {}).get('details', ''),
                'created_at': alert['created_at']
            }
            for page in pages
            for alert in page.json().get('alerts', [])
            if alert['status'] == 'triggered'
        ]

        # if more than one alert is listed for a check, only return the latest
        latest_alerts = []
        alerts.sort(key=lambda alert: alert['created_at'], reverse=True)
        for alert in alerts:
            alert_short = {'test': alert['test'], 'details': alert['details']}
            if alert_short not in latest_alerts:
                latest_alerts.append(alert_short)
        return latest_alerts

    def _trigger_alert(self, test_id, test_desc, msg, accreditation, details):
        report_links = self._get_report_links(
            test_desc, link_format='Report: {name}|{url}'
        )
        links = [
            {
                'text': t, 'href': rl
            } for t,
            rl in [rl.split('|') for rl in report_links]
        ]

        if test_desc['test'].test.runbook_url:
            links.append(
                {
                    'text': 'Runbook',
                    'href': test_desc['test'].test.runbook_url
                }
            )

        pagerduty.send_event(
            action='trigger',
            check=test_id,
            title=msg['title'],
            source=accreditation,
            severity='error',
            details=details,
            links=links,
            creds=self._creds
        )

    def _update_alert(self, test_id, test_desc, msg, accreditation, details):
        # NOTE: Unfortunately the PD API v2 doesn't have any way
        #       that I've found to actually update an existing alert
        #       or incident. A new log entry with the details is added
        #       but is hidden away. The visible info when you click on
        #       an incident remains as it was when it was first
        #       triggered.
        #       We can work around this by resolving the incident and
        #       then opening a new one.
        self._resolve_alert(test_id, test_desc, msg, accreditation)
        self._trigger_alert(test_id, test_desc, msg, accreditation, details)

    def _resolve_alert(self, test_id, test_desc, msg, accreditation):
        pagerduty.send_event(
            action='resolve',
            check=test_id,
            title=msg['title'],
            source=accreditation,
            severity='error',
            creds=self._creds
        )


class FindingsNotifier(_BaseNotifier):
    """
    Findings notifier class.

    Notifications are sent using the Findings API.  This notifier is
    configurable via :class:`compliance.config.ComplianceConfig`.
    """

    def __init__(self, results, controls, push_error=False):
        """
        Construct and initialize the Findings notifier object.

        :param results: dictionary generated by
          :py:class:`compliance.runners.CheckMode` at the end of the execution.
        :param controls: the control descriptor that manages accreditations.
        """
        super(FindingsNotifier, self).__init__(results, controls, push_error)
        self._config = get_config().get('notify.findings')
        self._creds = get_config().creds
        api_key = self._creds['findings'].api_key
        authenticator = IAMAuthenticator(apikey=api_key)
        self.findings_api = FindingsApiV1(authenticator=authenticator)

    def notify(self):
        """Send notifications to the Findings API."""
        if self._push_error:
            self.logger.error(
                'Remote locker push failed.  Findings notifier not triggered.'
            )
            return
        self.logger.info('Running the Findings notifier...')
        if not self._config:
            self.logger.warning('Using findings notification without config')

        messages = list(self._messages_by_accreditations().items())
        messages.sort(key=lambda x: x[0])
        for accreditation, desc in messages:
            if accreditation not in self._config:
                continue
            findings_api_endpoint = self._config[accreditation]
            self.findings_api.set_service_url(findings_api_endpoint)

            passed, failed, warned, errored = self._split_by_status(desc)
            for _, _, msg in (failed + errored + passed + warned):
                self._create_findings(msg['body'])

    def _create_findings(self, data):
        occurrence_list = data['occurrence_list']
        account_id = data['account_id']
        provider_id = data['provider_id']
        status = 0

        for occurrence in occurrence_list:
            try:
                response = self.findings_api.create_occurrence(
                    account_id=account_id,
                    provider_id=provider_id,
                    **occurrence
                )
                self.logger.info(response.status_code)
            except ApiException as e:
                status = e.code
                self.logger.error(
                    'Finding creation failed '
                    f'for occurrence id {occurrence["id"]} '
                    f'with {str(e.code)}: {str(e)}'
                )
            except Exception as e:
                status = -1
                self.logger.error(f'Unexpected error occurred: {str(e)}')
        return status


def get_notifiers():
    """
    Provide a dictionary of all notifier class objects.

    This dictionary contains all valid notifier choices for the ``--notify``
    option in the CLI as keys and their corresponding notifier classes as
    values.

    NOTE: When adding/removing a notifier, update this dictionary accordingly.
    """
    return {
        'stdout': FDNotifier,
        'slack': SlackNotifier,
        'pagerduty': PagerDutyNotifier,
        'gh_issues': GHIssuesNotifier,
        'locker': LockerNotifier,
        'findings': FindingsNotifier
    }



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compliance fetcher automation module."""

import os
import shutil
import tempfile
import unittest

from compliance.config import get_config
from compliance.utils.http import BaseSession

import requests


class ComplianceFetcher(unittest.TestCase):
    """Compliance fetcher automation TestCase class."""

    @classmethod
    def tearDownClass(cls):
        """Perform clean up."""
        if hasattr(cls, '_session'):
            cls._session.close()

    @classmethod
    def session(cls, url=None, creds=None, **headers):
        """
        Provide a requests session object with User-Agent header.

        :param url: optional base URL for the session requests to use.  A url
          argument triggers a new session object to be created whereas no url
          argument will return the current session object if one exists.
        :param creds: optional authentication credentials.
        :param headers: optional kwargs to add to session headers.

        :returns: a requests Session object.
        """
        if url is not None and hasattr(cls, '_session'):
            cls._session.close()
            delattr(cls, '_session')
        if not hasattr(cls, '_session'):
            if url:
                cls._session = BaseSession(url)
            else:
                cls._session = requests.Session()
            if creds:
                cls._session.auth = creds
            cls._session.headers.update(headers)
            org = cls.config.raw_config.get('org', {}).get('name', '')
            ua = f'{org.lower().replace(" ", "-")}-compliance-checks'
            cls._session.headers.update({'User-Agent': ua})
        return cls._session

    def __init__(self, *args, **kwargs):
        """Construct and initialize the fetcher test object."""
        super(ComplianceFetcher, self).__init__(*args, **kwargs)
        if hasattr(ComplianceFetcher, 'config'):
            self.config = ComplianceFetcher.config
        else:
            self.config = get_config()

    def fetchURL(self, url, params=None, creds=None):  # noqa: N802
        """
        Retrieve remote content through an HTTP GET request.

        Helper/Convenience method.

        :param url: the URL of the file.
        :param params: optional parameters to include in the GET request.
        :param creds: optional tuple with (user, password) to be used.

        :returns: the remote content.
        """
        org = self.config.raw_config.get('org', {}).get('name', '')
        ua = f'{org.lower().replace(" ", "-")}-compliance-checks'
        response = requests.get(
            url, params=params, auth=creds, headers={'User-Agent': ua}
        )
        response.raise_for_status()
        return response.content

    def fetchCloudantDoc(self, db_url, params=None):  # noqa: N802
        """
        Retrieve a Cloudant document.

        Helper/Convenience method.

        :param db_url: the URL to the Cloudant doc to retrieve.
        :param params: optional parameters to include in the GET request.

        :returns: the Cloudant document content.
        """
        creds = self.config.creds
        return self.fetchURL(
            db_url,
            params=params,
            creds=(creds['cloudant'].user, creds['cloudant'].password)
        )


def fetch(url, name):
    """
    Write content retrieved from provided url to a file.

    :param url: the URL to GET content from.
    :param name: the name of the file to write to in the TMPDIR.

    :returns: the path to the file.
    """
    r = requests.get(url)
    r.raise_for_status()
    path = os.path.join(tempfile.gettempdir(), name)

    with open(path, 'wb') as f:
        r.raw.decode_content = True
        shutil.copyfileobj(r.raw, f)

    return path



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compliance Pagerduty service helper."""

import json
from urllib.parse import urljoin

from compliance.utils.credentials import Config

import requests

PAGERDUTY_API_URL = 'https://api.pagerduty.com'
PD_EVENTS_V2_URL = 'https://events.pagerduty.com/v2/enqueue'
# 100 is the maximum page size
PAGES_LIMIT = 100


def _init_request(path, params, headers, creds):
    credentials = creds or Config()
    hdrs = {
        'Accept': 'application/vnd.pagerduty+json;version=2',
        'Authorization': f'Token token={credentials["pagerduty"].api_key}'
    }
    if headers:
        hdrs.update(headers)
    params = params or {}
    url = urljoin(PAGERDUTY_API_URL, path)
    return url, params, hdrs


def get(path, params=None, headers=None, creds=None):
    """
    Perform a GET operation.

    Returns the pages wrapped as :py:class:`requests.Response` object.
    This uses the PD API v2.

    :param path: API endpoint to call (e.g. 'users')
    :param params: a dictionary with parameters
    :param headers: a dictionary with headers to include
    :param creds: a Config object with PagerDuty credentials
    """
    url, params, hdrs = _init_request(path, params, headers, creds)
    offset = 0
    params.update({'limit': PAGES_LIMIT, 'offset': offset})
    more = True
    while more:
        r = requests.get(url, headers=hdrs, params=params)
        yield r
        more = r.json().get('more', False)
        if more:
            offset = offset + PAGES_LIMIT
            params.update({'offset': offset})


def delete(path, params=None, headers=None, creds=None):
    """
    Perform a DELETE operation.

    Returns the result as :py:class:`requests.Response` object.
    This uses the PD API v2.

    :param path: API endpoint to call (e.g. 'users')
    :param params: a dictionary with parameters
    :param headers: a dictionary with headers to include
    :param creds: a Config object with PagerDuty credentials
    """
    url, params, hdrs = _init_request(path, params, headers, creds)
    return requests.delete(url, headers=hdrs, params=params)


def put(path, params=None, headers=None, creds=None):
    """
    Perform a PUT operation.

    Return the result as :py:class:`requests.Response` object.
    This uses the PD API v2.

    :param path: API endpoint to call (e.g. 'users')
    :param params: a dictionary with parameters
    :param headers: a dictionary with headers to include
    :param creds: a Config object with PagerDuty credentials
    """
    url, params, hdrs = _init_request(path, params, headers, creds)
    return requests.put(url, headers=hdrs, params=params)


def post(path, params=None, headers=None, creds=None):
    """
    Perform a POST operation.

    Returns the result as :py:class:`requests.Response` objects.
    This uses the PD API v2.

    :param path: API endpoint to call (e.g. 'users')
    :param params: a dictionary with parameters
    :param headers: a dictionary with headers to include
    :param creds: a Config object with PagerDuty credentials
    """
    url, params, hdrs = _init_request(path, params, headers, creds)
    return requests.post(url, headers=hdrs, params=params)


def send_event(
    action,
    check,
    title,
    source,
    severity='error',
    details='',
    links=None,
    creds=None
):
    """Send an event to PD using the Events API."""
    credentials = creds or Config()

    msg = {
        'event_action': action,
        'routing_key': credentials['pagerduty'].events_integration_key,
        'dedup_key': check,
        'payload': {
            'summary': title,
            'source': source,
            'severity': severity,
            'custom_details': details
        },
        'links': links or []
    }
    headers = {'Content-Type': 'application/json'}

    response = requests.post(
        PD_EVENTS_V2_URL, headers=headers, data=json.dumps(msg)
    )
    response.raise_for_status()
    if response.json().get('status') != 'success':
        raise RuntimeError('PagerDuty Error: ' + response.json())



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compliance check controls automation module."""

import copy
import itertools
import json
import os
from collections import defaultdict


class ControlDescriptor(object):
    """
    Class abstraction for controls.json content.

    Used when processing controls.json content.
    """

    def __init__(self, dirs=None):
        """Construct and initialize the ControlDescriptor object."""
        self._controls = {}
        self._paths = []
        for d in dirs or ['.']:
            json_file = os.path.join(os.path.abspath(d), 'controls.json')
            if not os.path.isfile(json_file):
                continue
            self._controls.update(json.loads(open(json_file).read()))
            self._paths.append(json_file)

    @property
    def paths(self):
        """All absolute paths to ``controls.json`` file(s)."""
        return self._paths

    @property
    def as_dict(self):
        """Provide control descriptor content as a modifiable dictionary."""
        return copy.deepcopy(self._controls)

    @property
    def accred_checks(self):
        """Provide all checks by accreditation (key) as a dictionary."""
        if not hasattr(self, '_accred_checks'):
            self._accred_checks = defaultdict(set)
            for check in self._controls.keys():
                accreds = self.get_accreditations(check)
                for accred in accreds:
                    self._accred_checks[accred].add(check)
        return self._accred_checks

    def get_accreditations(self, test_path):
        """
        Provide the accreditation list for a given test_path.

        This works for original and simplified controls formats.

        Original: {'test_path': {'key': {'sub-key: ['accred1', 'accred2']}}}
        Simplified: {'test_path': ['accred1', 'accred2']}

        :param test_path: the Python path to the test.
          For example: ``package.category.checks.module.TestClass``
        """
        check_details = self._controls.get(test_path, {})
        if isinstance(check_details, list):
            return set(check_details)
        accreditations = [
            itertools.chain(*c.values()) for c in check_details.values()
        ]
        return set(itertools.chain(*accreditations))

    def is_test_included(self, test_path, accreditations):
        """
        Provide boolean result of whether a check is part of accreditations.

        :param test_path: the Python path to the test. For instance:
          ``package.accr1.TestClass`` or ``package.accr2.test_function``
        :param accreditations: list of accreditations names where ``test_path``
          may be included.
        """
        current_accreditations = self.get_accreditations(test_path)
        matched = set(accreditations).intersection(set(current_accreditations))
        return len(matched) != 0



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compliance automation flow management module."""

import inspect
import json
import os
import re
import sys
import time
import unittest
from collections import defaultdict

from compliance.check import ComplianceCheck
from compliance.config import get_config
from compliance.controls import ControlDescriptor
from compliance.fetch import ComplianceFetcher
from compliance.fix import Fixer
from compliance.locker import Locker
from compliance.notify import get_notifiers
from compliance.report import ReportBuilder
from compliance.utils.exceptions import LockerPushError
from compliance.utils.path import (
    CHECK_PREFIX, FETCH_PREFIX, get_toplevel_dirpath, load_evidences_modules
)


class _BaseRunner(object):
    """Base class for fetcher and check processing."""

    def __init__(self, opts, extra_opts):
        self.opts = opts
        self.extra_opts = extra_opts
        self.load_errors = set()

    def __enter__(self):
        self.init_config()

    def __exit__(self, typ, val, traceback):
        pass

    def init_config(self):
        """Initialize the framework configuration."""
        self._load_compliance_config()
        self._init_dirs()
        ComplianceFetcher.config = self.config
        ComplianceCheck.config = self.config

    def init_locker(self, ttl_tolerance=0):
        """
        Initialize the framework locker.

        :param ttl_tolerance: Evidence TTL tolerance in seconds
        """
        self.locker = self._create_the_locker(ttl_tolerance)
        self.locker.init()
        self.locker.logger_init_msgs()
        for path in self.opts.force:
            self.locker.forced_evidence.append(path)
        ComplianceFetcher.locker = self.locker
        ComplianceCheck.locker = self.locker

    def get_test_candidates(self, suite):
        """
        Provide the test cases from a test suite.

        :param suite: a TestSuite object

        :returns: generator iterable of test cases
        """
        for suite_test in suite:
            if unittest.suite._isnotsuite(suite_test):
                yield suite_test
                continue
            for test in self.get_test_candidates(suite_test):
                yield test

    def _load_compliance_config(self):
        creds_path = os.path.expanduser(self.opts.creds_path)
        if not os.path.isfile(creds_path):
            raise ValueError(f'{creds_path} does not exist.')
        self.config = get_config()
        self.config.creds_path = creds_path
        self.config.load(self.opts.compliance_config)

    def _init_dirs(self):
        self.dirs = set()
        for p in self.extra_opts:
            if os.path.isdir(p):
                self.dirs.add(get_toplevel_dirpath(os.path.abspath(p)))
            elif os.path.isfile(p):
                dirpath = get_toplevel_dirpath(os.path.abspath(p))
                if dirpath is None:
                    continue
                self.dirs.add(dirpath)
            else:
                self.dirs.add('.')

        if not self.dirs:
            self.dirs.add(os.path.abspath('.'))

        at_least_one_valid_dir = any(
            (get_toplevel_dirpath(d) is not None for d in self.dirs)
        )
        if not at_least_one_valid_dir:
            raise ValueError(
                'None of the paths provided are valid directories.  '
                'Please provide at least one directory containing a '
                '"controls.json" file.'
            )
        for d in self.dirs:
            load_evidences_modules(d)

    def _create_the_locker(self, ttl_tolerance):
        dirname = self.config.get('locker.dirname')
        mode = self.opts.evidence

        gitconfig = self.config.get('locker.gitconfig')
        if mode == 'local':
            return Locker(
                name=dirname, ttl_tolerance=ttl_tolerance, gitconfig=gitconfig
            )
        repo_url = self.config.get('locker.repo_url')
        if repo_url is None:
            raise ValueError(f'Evidence mode "{mode}" requires a URL.')
        return Locker(
            name=dirname,
            repo_url=repo_url,
            creds=self.config.creds,
            do_push=True if mode == 'full-remote' else False,
            ttl_tolerance=ttl_tolerance,
            gitconfig=gitconfig
        )


class FetchMode(_BaseRunner):
    """The fetcher process flow."""

    def __enter__(self):
        """Initialize fetcher mode processing."""
        super(FetchMode, self).__enter__()
        self.init_locker(self.config.get('locker.ttl_tolerance', 0))
        return self

    def __exit__(self, typ, val, traceback):
        """Handle post fetcher test execution processing."""
        super(FetchMode, self).__exit__(typ, val, traceback)
        # make sure that all added evidence are committed
        self.locker.checkin()

    def get_fetchers(self):
        """Provide all compliance framework fetcher classes."""
        fetchers = set()
        for loc in self.dirs:
            tl = unittest.TestLoader()
            tl.testMethodPrefix = FETCH_PREFIX
            candidates = self.get_test_candidates(
                tl.discover(loc, f'{FETCH_PREFIX}*.py')
            )
            for candidate in candidates:
                if issubclass(candidate.__class__, ComplianceFetcher):
                    fetchers.add(candidate.__class__)
            for load_err in tl.errors:
                try:
                    locate = re.search(
                        '^Failed to import test module: (.+?)\n.*?', load_err
                    )
                    if locate.group(1).split('.')[-1].startswith(FETCH_PREFIX):
                        self.load_errors.add(load_err)
                except AttributeError:
                    pass
        if not (self.opts.include or self.opts.exclude):
            return fetchers
        include = {f'{f.__module__}.{f.__name__}' for f in fetchers}
        if self.opts.include:
            include = set(json.loads(open(self.opts.include).read()))
        exclude = set()
        if self.opts.exclude:
            exclude = set(json.loads(open(self.opts.exclude).read()))
        return filter(
            lambda f: f'{f.__module__}.{f.__name__}' in include - exclude,
            fetchers
        )

    def run_fetchers(self, reruns=None):
        """
        Execute fetchers.

        :param reruns: A list of fetchers in dot notation to rerun

        :returns: Success (True) if no errors other than dependency unavailable
        """
        loader = unittest.TestLoader()
        loader.testMethodPrefix = FETCH_PREFIX
        fetchers = unittest.TestSuite()
        if reruns is None:
            fetcher_overrides = [
                fo for fo in self.extra_opts if not os.path.isdir(fo)
            ]
            if fetcher_overrides:
                fetchers.addTests(loader.loadTestsFromNames(fetcher_overrides))
            else:
                for fetcher in self.get_fetchers():
                    fetchers.addTests(loader.loadTestsFromTestCase(fetcher))
        else:
            self.config.dependency_rerun = True
            self.locker.reset_depenency_rerun()
            fetchers.addTests(loader.loadTestsFromNames(reruns))
        runner = unittest.TextTestRunner(
            verbosity=self.opts.verbose, resultclass=ComplianceBaseResult
        )
        return all(
            (
                'DependencyUnavailableError' in tb.split('Traceback')[-1]
                for (_, tb) in runner.run(fetchers).errors
            )
        )


class CheckMode(_BaseRunner):
    """The check process flow."""

    def __init__(self, opts, extra_opts):
        """
        Construct and initialize the check mode context manager.

        :param opts: arguments provided from the command line.
        :param extra_opts: additional arguments provided from the command line.
        """
        super(CheckMode, self).__init__(opts, extra_opts)
        self.accreds = [a.strip() for a in opts.check.split(',')]
        # Backward compatibility to support ghe_issues option
        self.notifiers = [
            n.strip().replace('ghe_issues', 'gh_issues')
            for n in opts.notify.split(',')
        ]
        self.push_error = False

    def __enter__(self):
        """Initialize check mode processing."""
        super(CheckMode, self).__enter__()
        self.init_locker()
        return self

    def __exit__(self, typ, val, traceback):
        """Handle post check test execution processing."""
        super(CheckMode, self).__exit__(typ, val, traceback)
        try:
            self.build_reports()
            # When in full-remote mode, fixers only run if push was successful
            self.fix_failures()
        except LockerPushError as lpe:
            self.locker.logger.error(str(lpe))
            self.push_error = True
        self.run_notifiers()

    def init_config(self):
        """Initialize the framework configuration for check execution."""
        super(CheckMode, self).init_config()
        self.results = None
        self.controls = ControlDescriptor(self.dirs)

    def get_checks(self):
        """Provide the appropriate compliance framework check classes."""
        checks = set()
        tests_found = set()
        for loc in self.dirs:
            tl = unittest.TestLoader()
            tl.testMethodPrefix = CHECK_PREFIX
            candidates = self.get_test_candidates(
                tl.discover(loc, f'{CHECK_PREFIX}*.py')
            )
            for test in [c.__class__ for c in candidates]:
                path = f'{test.__module__}.{test.__name__}'
                tests_found.add(path)
                in_accred_grouping = self.controls.is_test_included(
                    path, self.accreds
                )
                if issubclass(test, ComplianceCheck) and in_accred_grouping:
                    test.tests = [
                        method for method in dir(test) if (
                            method.startswith(CHECK_PREFIX) and (
                                inspect.ismethod(getattr(test, method))
                                or inspect.isfunction(getattr(test, method))
                            )
                        )
                    ]
                    checks.add(test)
            for load_err in tl.errors:
                try:
                    locate = re.search(
                        '^Failed to import test module: (.+?)\n.*?', load_err
                    )
                    for accred in self.accreds:
                        for check in self.controls.accred_checks[accred]:
                            if check.startswith(locate.group(1)):
                                self.load_errors.add(
                                    f'Unable to load {check}\n\n{load_err}'
                                )
                                tests_found.add(check)
                except AttributeError:
                    pass
        expected_checks = set()
        for accred, checks_in_accred in self.controls.accred_checks.items():
            if accred in self.accreds:
                expected_checks.update(checks_in_accred)
        for check_not_found in expected_checks - tests_found:
            self.load_errors.add(
                (
                    f'Unable to load {check_not_found}\n\n'
                    f'The check {check_not_found} was not found.  '
                    'Please validate that the path provided is correct.'
                )
            )
        return checks

    def run_checks(self):
        """
        Execute checks.

        :returns: Success (True) if no errors encountered
        """
        loader = unittest.TestLoader()
        loader.testMethodPrefix = CHECK_PREFIX
        checks = unittest.TestSuite()
        for check in self.get_checks():
            checks.addTests(loader.loadTestsFromTestCase(check))
        runner = unittest.TextTestRunner(
            verbosity=self.opts.verbose, resultclass=ComplianceCheckResult
        )
        check_run = runner.run(checks)
        self.results = check_run.results
        return False if check_run.errors else True

    def fix_failures(self):
        """Fix failures if fixer methods are included in checks."""
        if self.opts.fix != 'off':
            fixer = Fixer(self.results, dry_run=(self.opts.fix == 'dry-run'))
            fixer.fix()

    def build_reports(self):
        """Generate reports based on check results."""
        builder = ReportBuilder(self.locker, self.results, self.controls)
        builder.build()

    def run_notifiers(self):
        """Execute all requested notifiers."""
        sys.stdout.flush()
        sys.stderr.flush()
        notifiers = get_notifiers()
        for notifier_name in self.notifiers:
            notifier_args = [self.results, self.controls]
            if notifier_name == 'locker':
                notifier_args.append(self.locker)
            notifier = notifiers[notifier_name](
                *notifier_args, push_error=self.push_error
            )
            notifier.notify()


class ComplianceBaseResult(unittest.TextTestResult):
    """Base Compliance result class."""

    def startTest(self, test):  # noqa: N802
        """Start test timer for each test."""
        super(ComplianceBaseResult, self).startTest(test)
        if self.showAll:
            self.start_time = time.perf_counter()

    def stopTest(self, test):  # noqa: N802
        """Report on execution time at the end of each test."""
        super(ComplianceBaseResult, self).stopTest(test)
        if self.showAll:
            time_taken = time.perf_counter() - self.start_time
            self.stream.write(f'{self.getDescription(test)} - ran in: ')
            self.stream.writeln(f'{time_taken:.3f}s')
            self.stream.flush()


class ComplianceCheckResult(ComplianceBaseResult):
    """Compliance check result class."""

    def __init__(self, *args, **kwargs):
        """Construct and initialize the compliance check result."""
        super(ComplianceCheckResult, self).__init__(*args, **kwargs)
        self.results = defaultdict(dict)

    def addSuccess(self, test):  # noqa: N802
        """
        Add test successes and warnings to check results.

        :param test: a ``unittest.TestCase`` object.
        """
        super(ComplianceCheckResult, self).addSuccess(test)
        self.record(test, 'pass' if test.warnings_count() == 0 else 'warn')

    def addError(self, test, err):  # noqa: N802
        """
        Add test errors to check results.

        :param test: a ``unittest.TestCase`` object.
        :param err: a tuple of the form returned by sys.exc_info()
        """
        super(ComplianceCheckResult, self).addError(test, err)
        self.record(test, 'error')

    def addFailure(self, test, err):  # noqa: N802
        """
        Add test failures to check results.

        :param test: a ``unittest.TestCase`` object.
        :param err: a tuple of the form returned by sys.exc_info()
        """
        super(ComplianceCheckResult, self).addFailure(test, err)
        self.record(test, 'fail')

    def record(self, test, status):
        """
        Populate the results as expected by downstream reports and notifiers.

        :param test: a ``unittest.TestCase`` object.
        :param status: a string status (`pass`, `warn`, `fail`, or `error`)
        """
        self.results[test.id()] = {
            'status': status,
            'timestamp': time.time(),
            'test': ComplianceTestWrapper(test)
        }


class ComplianceTestWrapper(object):
    """
    TestCase wrapper class.

    Wraps the TestCase test object as an attribute to ensure backwards
    compatibility with checks, reporting, and notifiers.
    """

    def __init__(self, test):
        """Construct and initialize the test case wrapper object."""
        self.test = test



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compliance HTTP helpers."""

import requests


class BaseSession(requests.Session):
    """Subclass of requests.Session to support a base URL."""

    def __init__(self, baseurl):
        """
        Set the baseurl for the session.

        All requests using this session will be run against this URL.
        """
        super(BaseSession, self).__init__()
        self.baseurl = baseurl.strip('/')

    def prepare_request(self, request):
        """Prefix with self.baseurl if request.url does not include it."""
        if not request.url.startswith(self.baseurl):
            # the behavior of urljoin isn't what we want here
            request.url = f'{self.baseurl}/{request.url.lstrip("/")}'
        return super(BaseSession, self).prepare_request(request)
